import pandas as pd
import io

csv = """
Name,Sex,Count,Year
Noah,M,18252,2020
Julius,M,960,2020
Karen,M,6,2020
Karen,F,325,2020
Noah,F,305,2020
"""

csv2 = """
nyt_name,category
Karen,boomer
Julius,mythology
Karen,boomer2
Freya,mythology
"""

baby = pd.read_csv(io.StringIO(csv))
nyt = pd.read_csv(io.StringIO(csv2))["nyt_name"].rename(None)

# check that merging with unnamed series causes a pandas error but doesn't
# crash the backend
baby.merge(nyt)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
from functools import lru_cache
from typing import FrozenSet, Iterator, Mapping, Optional, Tuple, Union

from libcst._parser.conversions.expression import (
    convert_arg_assign_comp_for,
    convert_arglist,
    convert_argument,
    convert_atom,
    convert_atom_basic,
    convert_atom_curlybraces,
    convert_atom_ellipses,
    convert_atom_expr,
    convert_atom_expr_await,
    convert_atom_expr_trailer,
    convert_atom_parens,
    convert_atom_squarebrackets,
    convert_atom_string,
    convert_binop,
    convert_boolop,
    convert_comp_for,
    convert_comp_if,
    convert_comp_op,
    convert_comparison,
    convert_dictorsetmaker,
    convert_expression_input,
    convert_factor,
    convert_fstring,
    convert_fstring_content,
    convert_fstring_conversion,
    convert_fstring_equality,
    convert_fstring_expr,
    convert_fstring_format_spec,
    convert_lambda,
    convert_namedexpr_test,
    convert_not_test,
    convert_power,
    convert_sliceop,
    convert_star_arg,
    convert_star_expr,
    convert_subscript,
    convert_subscriptlist,
    convert_sync_comp_for,
    convert_test,
    convert_test_nocond,
    convert_test_or_expr_list,
    convert_testlist_comp_list,
    convert_testlist_comp_tuple,
    convert_trailer,
    convert_trailer_arglist,
    convert_trailer_attribute,
    convert_trailer_subscriptlist,
    convert_yield_arg,
    convert_yield_expr,
)
from libcst._parser.conversions.module import convert_file_input
from libcst._parser.conversions.params import (
    convert_argslist,
    convert_fpdef,
    convert_fpdef_assign,
    convert_fpdef_slash,
    convert_fpdef_star,
    convert_fpdef_starstar,
)
from libcst._parser.conversions.statement import (
    convert_annassign,
    convert_assert_stmt,
    convert_assign,
    convert_asyncable_funcdef,
    convert_asyncable_stmt,
    convert_augassign,
    convert_break_stmt,
    convert_classdef,
    convert_compound_stmt,
    convert_continue_stmt,
    convert_decorated,
    convert_decorator,
    convert_decorators,
    convert_del_stmt,
    convert_dotted_as_name,
    convert_dotted_as_names,
    convert_dotted_name,
    convert_except_clause,
    convert_expr_stmt,
    convert_for_stmt,
    convert_funcdef,
    convert_funcdef_annotation,
    convert_global_stmt,
    convert_if_stmt,
    convert_if_stmt_elif,
    convert_if_stmt_else,
    convert_import_as_name,
    convert_import_as_names,
    convert_import_from,
    convert_import_name,
    convert_import_relative,
    convert_import_stmt,
    convert_indented_suite,
    convert_nonlocal_stmt,
    convert_parameters,
    convert_pass_stmt,
    convert_raise_stmt,
    convert_return_stmt,
    convert_simple_stmt_line,
    convert_simple_stmt_partial,
    convert_simple_stmt_suite,
    convert_small_stmt,
    convert_stmt,
    convert_stmt_input,
    convert_suite,
    convert_try_stmt,
    convert_while_stmt,
    convert_with_item,
    convert_with_stmt,
)
from libcst._parser.conversions.terminals import (
    convert_ASYNC,
    convert_AWAIT,
    convert_DEDENT,
    convert_ENDMARKER,
    convert_FSTRING_END,
    convert_FSTRING_START,
    convert_FSTRING_STRING,
    convert_INDENT,
    convert_NAME,
    convert_NEWLINE,
    convert_NUMBER,
    convert_OP,
    convert_STRING,
)
from libcst._parser.parso.pgen2.generator import Grammar, generate_grammar
from libcst._parser.parso.python.token import PythonTokenTypes, TokenType
from libcst._parser.parso.utils import PythonVersionInfo, parse_version_string
from libcst._parser.production_decorator import get_productions
from libcst._parser.types.config import AutoConfig
from libcst._parser.types.conversions import NonterminalConversion, TerminalConversion
from libcst._parser.types.production import Production

# Keep this sorted alphabetically
_TERMINAL_CONVERSIONS_SEQUENCE: Tuple[TerminalConversion, ...] = (
    convert_DEDENT,
    convert_ENDMARKER,
    convert_INDENT,
    convert_NAME,
    convert_NEWLINE,
    convert_NUMBER,
    convert_OP,
    convert_STRING,
    convert_FSTRING_START,
    convert_FSTRING_END,
    convert_FSTRING_STRING,
    convert_ASYNC,
    convert_AWAIT,
)

# Try to match the order of https://docs.python.org/3/reference/grammar.html
_NONTERMINAL_CONVERSIONS_SEQUENCE: Tuple[NonterminalConversion, ...] = (
    convert_file_input,
    convert_stmt_input,  # roughly equivalent to single_input
    convert_expression_input,  # roughly equivalent to eval_input
    convert_stmt,
    convert_simple_stmt_partial,
    convert_simple_stmt_line,
    convert_simple_stmt_suite,
    convert_small_stmt,
    convert_expr_stmt,
    convert_annassign,
    convert_augassign,
    convert_assign,
    convert_pass_stmt,
    convert_continue_stmt,
    convert_break_stmt,
    convert_del_stmt,
    convert_import_stmt,
    convert_import_name,
    convert_import_relative,
    convert_import_from,
    convert_import_as_name,
    convert_dotted_as_name,
    convert_import_as_names,
    convert_dotted_as_names,
    convert_dotted_name,
    convert_return_stmt,
    convert_raise_stmt,
    convert_global_stmt,
    convert_nonlocal_stmt,
    convert_assert_stmt,
    convert_compound_stmt,
    convert_if_stmt,
    convert_if_stmt_elif,
    convert_if_stmt_else,
    convert_while_stmt,
    convert_for_stmt,
    convert_try_stmt,
    convert_except_clause,
    convert_with_stmt,
    convert_with_item,
    convert_asyncable_funcdef,
    convert_funcdef,
    convert_classdef,
    convert_decorator,
    convert_decorators,
    convert_decorated,
    convert_asyncable_stmt,
    convert_parameters,
    convert_argslist,
    convert_fpdef_slash,
    convert_fpdef_star,
    convert_fpdef_starstar,
    convert_fpdef_assign,
    convert_fpdef,
    convert_funcdef_annotation,
    convert_suite,
    convert_indented_suite,
    convert_namedexpr_test,
    convert_test,
    convert_test_nocond,
    convert_lambda,
    convert_boolop,
    convert_not_test,
    convert_comparison,
    convert_comp_op,
    convert_star_expr,
    convert_binop,
    convert_factor,
    convert_power,
    convert_atom_expr,
    convert_atom_expr_await,
    convert_atom_expr_trailer,
    convert_trailer,
    convert_trailer_attribute,
    convert_trailer_subscriptlist,
    convert_subscriptlist,
    convert_subscript,
    convert_sliceop,
    convert_trailer_arglist,
    convert_atom,
    convert_atom_basic,
    convert_atom_parens,
    convert_atom_squarebrackets,
    convert_atom_curlybraces,
    convert_atom_string,
    convert_fstring,
    convert_fstring_content,
    convert_fstring_conversion,
    convert_fstring_equality,
    convert_fstring_expr,
    convert_fstring_format_spec,
    convert_atom_ellipses,
    convert_testlist_comp_tuple,
    convert_testlist_comp_list,
    convert_test_or_expr_list,
    convert_dictorsetmaker,
    convert_arglist,
    convert_argument,
    convert_arg_assign_comp_for,
    convert_star_arg,
    convert_sync_comp_for,
    convert_comp_for,
    convert_comp_if,
    convert_yield_expr,
    convert_yield_arg,
)


def get_grammar_str(version: PythonVersionInfo, future_imports: FrozenSet[str]) -> str:
    """
    Returns an BNF-like grammar text that `parso.pgen2.generator.generate_grammar` can
    handle.

    While you should generally use `get_grammar` instead, this can be useful for
    debugging the grammar.
    """
    lines = []
    for p in get_nonterminal_productions(version, future_imports):
        lines.append(str(p))
    return "\n".join(lines) + "\n"


# TODO: We should probably provide an on-disk cache like parso and lib2to3 do. Because
# of how we're defining our grammar, efficient cache invalidation is harder, though not
# impossible.
@lru_cache()
def get_grammar(
    version: PythonVersionInfo,
    future_imports: Union[FrozenSet[str], AutoConfig],
) -> "Grammar[TokenType]":
    if isinstance(future_imports, AutoConfig):
        # For easier testing, if not provided assume no __future__ imports
        future_imports = frozenset(())
    return generate_grammar(get_grammar_str(version, future_imports), PythonTokenTypes)


@lru_cache()
def get_terminal_conversions() -> Mapping[str, TerminalConversion]:
    """
    Returns a mapping from terminal type name to the conversion function that should be
    called by the parser.
    """
    return {
        # pyre-fixme[16]: Optional type has no attribute `group`.
        re.match("convert_(.*)", fn.__name__).group(1): fn
        for fn in _TERMINAL_CONVERSIONS_SEQUENCE
    }


@lru_cache()
def validate_grammar() -> None:
    for fn in _NONTERMINAL_CONVERSIONS_SEQUENCE:
        fn_productions = get_productions(fn)
        if all(p.name == fn_productions[0].name for p in fn_productions):
            # all the production names are the same, ensure that the `convert_` function
            # is named correctly
            production_name = fn_productions[0].name
            expected_name = f"convert_{production_name}"
            if fn.__name__ != expected_name:
                raise Exception(
                    f"The conversion function for '{production_name}' "
                    + f"must be called '{expected_name}', not '{fn.__name__}'."
                )


def _get_version_comparison(version: str) -> Tuple[str, PythonVersionInfo]:
    if version[:2] in (">=", "<=", "==", "!="):
        return (version[:2], parse_version_string(version[2:].strip()))
    if version[:1] in (">", "<"):
        return (version[:1], parse_version_string(version[1:].strip()))
    raise Exception(f"Invalid version comparison specifier '{version}'")


def _compare_versions(
    requested_version: PythonVersionInfo,
    actual_version: PythonVersionInfo,
    comparison: str,
) -> bool:
    if comparison == ">=":
        return actual_version >= requested_version
    if comparison == "<=":
        return actual_version <= requested_version
    if comparison == "==":
        return actual_version == requested_version
    if comparison == "!=":
        return actual_version != requested_version
    if comparison == ">":
        return actual_version > requested_version
    if comparison == "<":
        return actual_version < requested_version
    raise Exception(f"Invalid version comparison specifier '{comparison}'")


def _should_include(
    requested_version: Optional[str], actual_version: PythonVersionInfo
) -> bool:
    if requested_version is None:
        return True
    for version in requested_version.split(","):
        comparison, parsed_version = _get_version_comparison(version.strip())
        if not _compare_versions(parsed_version, actual_version, comparison):
            return False
    return True


def _should_include_future(
    future: Optional[str],
    future_imports: FrozenSet[str],
) -> bool:
    if future is None:
        return True
    if future[:1] == "!":
        return future[1:] not in future_imports
    return future in future_imports


def get_nonterminal_productions(
    version: PythonVersionInfo, future_imports: FrozenSet[str]
) -> Iterator[Production]:
    for conversion in _NONTERMINAL_CONVERSIONS_SEQUENCE:
        for production in get_productions(conversion):
            if not _should_include(production.version, version):
                continue
            if not _should_include_future(production.future, future_imports):
                continue
            yield production


@lru_cache()
def get_nonterminal_conversions(
    version: PythonVersionInfo,
    future_imports: FrozenSet[str],
) -> Mapping[str, NonterminalConversion]:
    """
    Returns a mapping from nonterminal production name to the conversion function that
    should be called by the parser.
    """
    conversions = {}
    for fn in _NONTERMINAL_CONVERSIONS_SEQUENCE:
        for fn_production in get_productions(fn):
            if not _should_include(fn_production.version, version):
                continue
            if not _should_include_future(fn_production.future, future_imports):
                continue
            if fn_production.name in conversions:
                raise Exception(
                    f"Found duplicate '{fn_production.name}' production in grammar"
                )
            conversions[fn_production.name] = fn

    return conversions



import pandas as pd
import io

csv = '''
breed,type,longevity,size
Labrador Retriever,sporting,12.04,medium
German Shepherd,herding,9.73,large
Beagle,hound,12.3,small
Golden,sporting,12.04,medium
Yorkshire,toy,12.6,small
Bulldog,non-sporting,6.29,medium
'''

dogs = (pd.read_csv(io.StringIO(csv))
  [['breed', 'type', 'longevity']]
)

dogs.drop(index=1, axis=1)


# https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.stack.html
import pandas as pd
multicol1 = pd.MultiIndex.from_tuples([('weight', 'kg'),
                                       ('weight', 'pounds')])
df_multi_level_cols1 = pd.DataFrame([[1, 2], [2, 4]],
                                    index=['cat', 'dog'],
                                    columns=multicol1)
df_multi_level_cols1.stack()


import pandas as pd
import io

building_records_1844 = pd.read_csv(io.StringIO('''
building,location,established
Grande Hotel,4-5,1831
Jone's Farm,1-2,1842
Public Library,6-4,1836
Marietta House,1-7,1823
''')).set_index(['building', 'location'])

building_records_2020 = pd.read_csv(io.StringIO('''
building,location,established
Sam's Bakery,5-1,1962
Grande Hotel,4-5,1830
Public Library,6-4,1835
Mayberry's Factory,3-2,1924
''')).set_index(['building', 'location'])

building_records_1844.join(
    building_records_2020,
    how='inner',
    rsuffix='_2000',
)


import os
from pathlib import Path
import unittest

from pandas_tutor.__main__ import make_tutor_spec

test_cases = Path(__file__).parent / "e2e_golden"


class TestEndToEnd(unittest.TestCase):
    pass


def make_test_case(test_name):
    in_file = test_cases / f"{test_name}.py"
    golden_file = test_cases / f"{test_name}.py.golden"

    def test(self: TestEndToEnd):
        self.assertTrue(in_file.exists())
        self.assertTrue(golden_file.exists())

        code = in_file.read_text()

        # we'll just compare JSON strings here since we only apply special
        # encoding rules (e.g. NaN to None) after converting to JSON.
        res = make_tutor_spec(code)
        golden_res = golden_file.read_text().strip()

        self.assertEqual(res, golden_res)

    return test


# make all test cases dynamically!
for test_name in test_cases.iterdir():
    if test_name.suffix == ".py" and test_name.name != '__init__.py':
        if os.environ.get("CI", False) and test_name.stem.startswith("plot_"):
            # skip plotting test cases since gzipping isn't deterministic
            continue
        test_case = make_test_case(test_name.stem)
        setattr(TestEndToEnd, f"test_{test_name.stem}", test_case)



"""
serializes run.py outputs into json.
"""

from __future__ import annotations

import types
from typing import Any, List, Tuple, TypeVar, Union

import pandas as pd

from pandas_tutor.parse_nodes import JoinCall, MergeCall, StartOfChain

from . import util
from .diagram import (
    DataPair,
    DataSpec,
    DFSpec,
    DataTwoLHS,
    Diagram,
    ErrorOutput,
    Explanation,
    Group,
    GroupBySpec,
    GroupData,
    ImageSpec,
    Index,
    PrevRHS,
    RuntimeErrorInChain,
    RuntimeErrorInSetup,
    SeriesGroupBySpec,
    SeriesSpec,
    SyntaxErrorOutput,
    UnhandledData,
)
from .marks import make_marks
from .run import (
    EvalResult,
    ImageResult,
    RuntimeErrorResult,
    SyntaxErrorResult,
)

T = TypeVar("T")


def serialize(results: List[EvalResult]) -> Explanation:
    if len(results) == 0:
        return []

    # stop if results use too much memory
    total_mem_used = sum(util.mem_used(result.val) for result in results)
    if total_mem_used > util.MEM_LIMIT:
        result = results[-1]
        return [
            RuntimeErrorInChain(
                code_step=result.step.code,
                message=util.too_much_mem_msg(total_mem_used),
                fragment=result.fragment,
            )
        ]

    if len(results) == 1:
        # happens when user inputs `df` without a function call, or when
        # error happens in setup code
        return serialize_single(results[0])

    return [serialize_pair(before, after) for before, after in pairs(results)]


def serialize_single(result: EvalResult) -> Explanation:
    if isinstance(result, SyntaxErrorResult):
        return [SyntaxErrorOutput.from_parse_syntax_error(result.step)]
    elif isinstance(result, RuntimeErrorResult):
        return [RuntimeErrorInSetup.from_runtime_error_result(result)]
    return [
        Diagram(
            type=result.step.type_,
            code_step=result.step.code,
            fragment=result.fragment,
            marks=[],
            data=DataPair(lhs=serialize_step_val(result), rhs="no_rhs"),
        )
    ]


def serialize_pair(
    before: EvalResult, after: EvalResult
) -> Union[Diagram, ErrorOutput]:
    if isinstance(after, RuntimeErrorResult):
        return RuntimeErrorInChain.from_runtime_error_result(after)
    step = after.step

    marks = make_marks(step, before, after)

    lhs: Union[DataSpec, PrevRHS] = (
        serialize_step_val(before)
        if isinstance(before.step, StartOfChain)
        else "prev_rhs"
    )
    rhs = serialize_step_val(after)

    # HACK: special case for merge and join
    data: Union[DataTwoLHS, DataPair]
    if isinstance(step, (MergeCall, JoinCall)):
        lhs2 = (
            after.args["right"]
            if "right" in after.args
            else after.args["other"]
            if "other" in after.args
            else None
        )
        data = DataTwoLHS(
            lhs=lhs,
            lhs2=serialize_pd_val(lhs2),
            rhs=rhs,
        )
    else:
        data = DataPair(lhs=lhs, rhs=rhs)

    return Diagram(
        type=step.type_,
        code_step=step.code,
        fragment=after.fragment,
        marks=marks,
        data=data,
    )


def serialize_step_val(step: EvalResult) -> DataSpec:
    """
    if we have an EvalResult, use this for serializing so that we can handle
    special results like images
    """
    if isinstance(step, ImageResult):
        return serialize_image(step.val)
    return serialize_pd_val(step.val)


def serialize_pd_val(val: Any) -> DataSpec:
    if isinstance(val, pd.DataFrame):
        return DFSpec.from_pd(val)
    if isinstance(val, pd.Series):
        return SeriesSpec.from_pd(val)
    if isinstance(val, util.DataFrameGroupBy):
        return serialize_groupby(val)
    if isinstance(val, util.SeriesGroupBy):
        return serialize_seriesgroupby(val)

    if isinstance(val, types.ModuleType):
        # take off module path from the module output, otherwise tests
        # don't work in CI
        data = f"<module '{val.__name__}'>"
    else:
        data = repr(val)
    return UnhandledData(data=data)


def serialize_groupby(val: util.DataFrameGroupBy) -> GroupBySpec:
    col_names = util.grouping_labels(val)

    df_groups = util.get_groups(val)
    groups = [
        Group(
            name=list(name) if util.is_list_like(name) else [name],
            labels=labels.tolist(),
        )
        for name, labels in df_groups.items()
    ]

    df = util.ungroup(val)
    group_data = GroupData(columns=col_names, groups=groups)
    return GroupBySpec(
        columns=Index.from_pd(df.columns),
        index=Index.from_pd(df.index),
        data=util.df_data(df),
        group_data=group_data,
    )


def serialize_seriesgroupby(val: util.SeriesGroupBy) -> SeriesGroupBySpec:
    col_names = util.grouping_labels(val)

    df_groups = util.get_groups(val)
    groups = [
        Group(
            name=list(name) if util.is_list_like(name) else [name],
            labels=labels.tolist(),
        )
        for name, labels in df_groups.items()
    ]

    series = util.ungroup(val)
    group_data = GroupData(columns=col_names, groups=groups)
    return SeriesGroupBySpec(
        index=Index.from_pd(series.index),
        data=util.series_data(series),
        group_data=group_data,
    )


def serialize_image(val: Any) -> ImageSpec:
    return ImageSpec(util.base64_encode_plot(val))


def pairs(seq: List[T]) -> List[Tuple[T, T]]:
    return [(seq[i], seq[i + 1]) for i in range(len(seq) - 1)]



# Copyright 2004-2005 Elemental Security, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

# Modifications:
# Copyright David Halter and Contributors
# Modifications are dual-licensed: MIT and PSF.
# 99% of the code is different from pgen2, now.

# A fork of `parso.parser`.
# https://github.com/davidhalter/parso/blob/v0.3.4/parso/parser.py
#
# The following changes were made:
# - Typing was added.
# - Error recovery is removed.
# - The Jedi-specific _allowed_transition_names_and_token_types API is removed.
# - Improved error messages by using our exceptions module.
# - node_map/leaf_map were removed in favor of just calling convert_*.
# - convert_node/convert_leaf were renamed to convert_nonterminal/convert_terminal
# - convert_nonterminal is called regardless of the number of children. Parso avoids
#   calling it in some cases to avoid creating extra nodes.
# - The parser is constructed with the tokens to allow us to track a bit more state. As
#   As a consequence parser may only be used once.
# - Supports our custom Token class, instead of `parso.python.tokenize.Token`.


from dataclasses import dataclass, field
from typing import Generic, Iterable, List, Sequence, TypeVar, Union

from libcst._exceptions import (
    EOFSentinel,
    ParserSyntaxError,
    PartialParserSyntaxError,
    get_expected_str,
)
from libcst._parser.parso.pgen2.generator import DFAState, Grammar, ReservedString
from libcst._parser.parso.python.token import TokenType
from libcst._parser.types.token import Token

_NodeT = TypeVar("_NodeT")
_TokenTypeT = TypeVar("_TokenTypeT", bound=TokenType)
_TokenT = TypeVar("_TokenT", bound=Token)


@dataclass(frozen=False)
class StackNode(Generic[_TokenTypeT, _NodeT]):
    dfa: "DFAState[_TokenTypeT]"
    nodes: List[_NodeT] = field(default_factory=list)

    @property
    def nonterminal(self) -> str:
        return self.dfa.from_rule


def _token_to_transition(
    grammar: "Grammar[_TokenTypeT]", type_: _TokenTypeT, value: str
) -> Union[ReservedString, _TokenTypeT]:
    # Map from token to label
    if type_.contains_syntax:
        # Check for reserved words (keywords)
        try:
            return grammar.reserved_syntax_strings[value]
        except KeyError:
            pass

    return type_


# TODO: This should be an ABC, but there's a metaclass conflict between Generic and ABC
# that's fixed in Python 3.7.
class BaseParser(Generic[_TokenT, _TokenTypeT, _NodeT]):
    """Parser engine.

    A Parser instance contains state pertaining to the current token
    sequence, and should not be used concurrently by different threads
    to parse separate token sequences.

    See python/tokenize.py for how to get input tokens by a string.
    """

    tokens: Iterable[_TokenT]
    lines: Sequence[str]  # used when generating parse errors
    _pgen_grammar: "Grammar[_TokenTypeT]"
    stack: List[StackNode[_TokenTypeT, _NodeT]]
    # Keep track of if parse was called. Because a parser may keep global mutable state,
    # each BaseParser instance should only be used once.
    __was_parse_called: bool

    def __init__(
        self,
        *,
        tokens: Iterable[_TokenT],
        lines: Sequence[str],
        pgen_grammar: "Grammar[_TokenTypeT]",
        start_nonterminal: str,
    ) -> None:
        self.tokens = tokens
        self.lines = lines
        self._pgen_grammar = pgen_grammar
        first_dfa = pgen_grammar.nonterminal_to_dfas[start_nonterminal][0]
        self.stack = [StackNode(first_dfa)]
        self.__was_parse_called = False

    def parse(self) -> _NodeT:
        # Ensure that we don't re-use parsers.
        if self.__was_parse_called:
            raise Exception("Each parser object may only be used to parse once.")
        self.__was_parse_called = True

        for token in self.tokens:
            self._add_token(token)

        while True:
            tos = self.stack[-1]
            if not tos.dfa.is_final:
                expected_str = get_expected_str(
                    EOFSentinel.EOF, tos.dfa.transitions.keys()
                )
                raise ParserSyntaxError(
                    f"Incomplete input. {expected_str}",
                    lines=self.lines,
                    raw_line=len(self.lines),
                    raw_column=len(self.lines[-1]),
                )

            if len(self.stack) > 1:
                self._pop()
            else:
                return self.convert_nonterminal(tos.nonterminal, tos.nodes)

    def convert_nonterminal(
        self, nonterminal: str, children: Sequence[_NodeT]
    ) -> _NodeT:
        ...

    def convert_terminal(self, token: _TokenT) -> _NodeT:
        ...

    def _add_token(self, token: _TokenT) -> None:
        """
        This is the only core function for parsing. Here happens basically
        everything. Everything is well prepared by the parser generator and we
        only apply the necessary steps here.
        """
        grammar = self._pgen_grammar
        stack = self.stack
        # pyre-fixme[6]: Expected `_TokenTypeT` for 2nd param but got `TokenType`.
        transition = _token_to_transition(grammar, token.type, token.string)

        while True:
            try:
                plan = stack[-1].dfa.transitions[transition]
                break
            except KeyError:
                if stack[-1].dfa.is_final:
                    try:
                        self._pop()
                    except PartialParserSyntaxError as ex:
                        # Upconvert the PartialParserSyntaxError to a ParserSyntaxError
                        # by backfilling the line/column information.
                        raise ParserSyntaxError(
                            ex.message,
                            lines=self.lines,
                            raw_line=token.start_pos[0],
                            raw_column=token.start_pos[1],
                        )
                    except Exception as ex:
                        # convert_nonterminal may fail due to a bug in our code. Try to
                        # recover enough to at least tell us where in the file it
                        # failed.
                        raise ParserSyntaxError(
                            f"Internal error: {ex}",
                            lines=self.lines,
                            raw_line=token.start_pos[0],
                            raw_column=token.start_pos[1],
                        )
                else:
                    # We never broke out -- EOF is too soon -- Unfinished statement.
                    #
                    # BUG: The `expected_str` may not be complete because we already
                    # popped the other possibilities off the stack at this point, but
                    # it still seems useful to list some of the possibilities that we
                    # could've expected.
                    expected_str = get_expected_str(
                        token, stack[-1].dfa.transitions.keys()
                    )
                    raise ParserSyntaxError(
                        f"Incomplete input. {expected_str}",
                        lines=self.lines,
                        raw_line=token.start_pos[0],
                        raw_column=token.start_pos[1],
                    )
            except IndexError:
                # I don't think this will ever happen with Python's grammar, because if
                # there are any extra tokens at the end of the input, we'll instead
                # complain that we expected ENDMARKER.
                #
                # However, let's leave it just in case.
                expected_str = get_expected_str(token, EOFSentinel.EOF)
                raise ParserSyntaxError(
                    f"Too much input. {expected_str}",
                    lines=self.lines,
                    raw_line=token.start_pos[0],
                    raw_column=token.start_pos[1],
                )

        # Logically, `plan` is always defined, but pyre can't reasonably determine that.
        # pyre-fixme[18]: Global name `plan` is undefined.
        stack[-1].dfa = plan.next_dfa

        for push in plan.dfa_pushes:
            stack.append(StackNode(push))

        leaf = self.convert_terminal(token)
        stack[-1].nodes.append(leaf)

    def _pop(self) -> None:
        tos = self.stack.pop()
        # Unlike parso and lib2to3, we call `convert_nonterminal` unconditionally
        # instead of only when we have more than one child. This allows us to create a
        # far more consistent and predictable tree.
        new_node = self.convert_nonterminal(tos.dfa.from_rule, tos.nodes)
        self.stack[-1].nodes.append(new_node)



import pandas as pd
import io

csv = """
Name,Sex,Count,Year
Noah,M,18252,2020
Julius,M,960,2020
Karen,M,6,2020
Noah,F,305,2020
"""

baby = pd.read_csv(io.StringIO(csv)).set_index("Name")

baby.merge(baby)



# https://pandas.pydata.org/pandas-docs/stable/user_guide/reshaping.html#reshaping-by-pivoting-dataframe-objects
import io
import warnings

import pandas as pd

# getting a weird warning that only shows up in unittest, so let's silence it
warnings.filterwarnings("ignore", category=DeprecationWarning)

csv = """
foo,bar,baz,zoo
one,A,1,x
one,B,2,y
one,C,3,z
two,A,4,q
two,B,5,w
two,C,6,t
"""

df = pd.read_csv(io.StringIO(csv))

# produces an empty dataframe with two row labels
df.pivot(index=["foo"], columns=["bar", "baz", "zoo"])



"""
parses a pandas snippet using libcst. the important thing here is to get the
positions of each expression within the snippet so that we can selectively run
parts of it. there are a bunch of hard-coded heuristics so that each step in
the result hopefully corresponds to one new dataframe.

in the future it would be ideal to hook into bdb directly like python tutor
does. that way we step through the code itself, and it'd be a lot easier to
integrate into python tutor's existing backend.

right now i'm not sure how to make bdb step through individual function calls
in a chain. it doesn't seem to be the default as this pytutor link shows:
https://pythontutor.com/visualize.html#code=s%20%3D%20'hello%20world%20!!%20'%0Atest%20%3D%20s.strip%28%29.replace%28'%20',%20'!'%29.split%28'!'%29%0Aprint%28test%29&cumulative=false&curInstr=2&heapPrimitives=nevernest&mode=display&origin=opt-frontend.js&py=3&rawInputLstJSON=%5B%5D&textReferences=false
"""

# For forward type references: https://stackoverflow.com/a/33533514
from __future__ import annotations

import typing as t
from warnings import warn

import libcst as cst
import libcst.matchers as m
import libcst.metadata as cstm

from pandas_tutor import util
from pandas_tutor.util import CodePosition, CodeRange

from .parse_nodes import (
    AggCall,
    ApplyCall,
    AssignCall,
    Axis,
    BoolExprStep,
    Call,
    ChainStatement,
    ChainStep,
    DropCall,
    GetCall,
    GroupByCall,
    HeadCall,
    JoinCall,
    MeltCall,
    MergeCall,
    ParsedModule,
    ParseResult,
    ParseSyntaxError,
    PassThroughCall,
    PivotCall,
    PivotTableCall,
    RawCode,
    RenameCall,
    ResetIndexCall,
    SetIndexCall,
    SortValuesCall,
    StackCall,
    StartOfChain,
    SubsComparison,
    Subscript,
    SubscriptEl,
    SubsEval,
    SubsSlice,
    TailCall,
    UnstackCall,
    VerbatimStatement,
)

T = t.TypeVar("T")


def parse(code: str) -> ParseResult:
    try:
        tree = cst.parse_module(code)
    except cst.ParserSyntaxError as e:
        pos = CodePosition(e.editor_line - 1, e.editor_column)
        return ParseSyntaxError(
            code=RawCode(code), error_msg=str(e), location=CodeRange(pos, pos)
        )

    with_meta = cstm.MetadataWrapper(tree)
    positions = with_meta.resolve(cstm.PositionProvider)
    sam = PandasParser(tree, positions)
    _ = with_meta.visit(sam)
    return sam.root


def parse_as_json(code: str) -> str:
    node = parse(code)
    return ParsedModule.to_json(node)


# parse expressions and assignments into chains
#
# NOTE: need to update this if we want handle annotated assignments
# (x: int = 5) or augmenting assignments (x += 5)
is_chain_stmt = m.Expr() | m.Assign()

# Used to ensure that we don't recurse into nested calls / subscripts
is_argument = m.Arg() | m.SubscriptElement()

# we also shouldn't recurse into boolean expressions, like:
#
# s1 | s2
# df['a'] < 5
is_bool_expr = m.BinaryOperation() | m.Comparison()

is_attribute_call = m.Call(func=m.Attribute())


def matches(call_cls: t.Type[Call]) -> m.Call:
    # returns libcst matcher from a parse_node.Call subclass
    return m.Call(func=m.Attribute(attr=m.Name(call_cls.fn_name)))


# set of all parsed function calls
parsed_calls = {cls.fn_name for cls in Call.__subclasses__()}


def fn_name(call: cst.Call):
    func = t.cast(cst.Attribute, call.func)
    return func.attr.value


def is_parsed_call(call: cst.Call) -> bool:
    return fn_name(call) in parsed_calls


is_loc_iloc = m.Subscript(
    value=m.Attribute(attr=m.Name("loc") | m.Name("iloc"))
)

is_boolean_slice = m.Comparison() | m.BinaryOperation(
    operator=(m.BitOr() | m.BitAnd())
)


def get_arg(
    args: t.Sequence[cst.Arg], position: int, keyword: t.Optional[str] = None
) -> t.Optional[cst.Arg]:
    """
    some function args can be passed by both position and keyword. if the arg
    is passed by keyword, it should have priority over the position, so:

        df.sort_values('Name')
        df.sort_values(ascending=False, by='Name')

    should all get the Arg for 'Name'
    """
    if keyword is not None:
        for arg in args:
            if arg.keyword is not None and arg.keyword.value == keyword:
                return arg
    if position >= len(args):
        return None

    arg = args[position]
    return arg if arg.keyword is None else None


def make_axis(value: str) -> Axis:
    return (
        "columns"
        if value == "1" or "columns" in value
        else "index"
        if value == "0" or "index" in value
        # index is the default for most pandas methods so we'll just fall
        # back to that...maybe we should raise an error in this case instead
        else "index"
    )


class ParserBase(m.MatcherDecoratableVisitor):
    """base class for our libcst node visitors"""

    cst_root: cst.Module

    # We need to initialize with this rather than use metadata since subclasses
    # will visit subtrees of the module and metadata deps only work on
    # module-level visitors.
    positions: t.Mapping[cst.CSTNode, cstm.CodeRange]

    def __init__(
        self,
        cst_root: cst.Module,
        positions: t.Mapping[cst.CSTNode, cstm.CodeRange],
    ):
        super().__init__()
        self.cst_root = cst_root
        self.positions = positions
        self.__post_init__()

    def __post_init__(self):
        """called after __init__() to let subclasses initialize other fields"""
        pass

    @classmethod
    def from_parent(cls, parent: ParserBase):
        return cls(parent.cst_root, parent.positions)

    def code_for(self, cst_node):
        return RawCode(self.cst_root.code_for_node(cst_node))

    def location(self, cst_node) -> CodeRange:
        code_range = self.positions[cst_node]
        # subtract 1 from line to make everything 0-indexed
        start = CodePosition(
            line=code_range.start.line - 1, ch=code_range.start.column
        )
        end = CodePosition(
            line=code_range.end.line - 1, ch=code_range.end.column
        )
        return CodeRange(start, end)

    def make_node(
        self,
        cls: t.Type[T],
        cst_node: cst.CSTNode,
        code=None,
        location=None,
        **kwargs,
    ) -> T:
        if code is None:
            code = self.code_for(cst_node)
        if location is None:
            location = self.location(cst_node)
        return cls(code=code, location=location, **kwargs)  # type: ignore


class PandasParser(ParserBase):
    """top level visitor that handles statements"""

    root: ParsedModule

    def visit_Module(self, cst_node: cst.Module):
        statements = util.flatmap(self.parse_statements, cst_node.body)
        self.root = self.make_node(
            ParsedModule, cst_node, statements=list(statements)
        )

    def parse_statements(
        self,
        stmts: t.Union[cst.BaseCompoundStatement, cst.SimpleStatementLine],
    ) -> t.List[t.Union[VerbatimStatement, ChainStatement]]:
        if isinstance(stmts, cst.BaseCompoundStatement):
            return [self.make_verbatim_stmt(stmts)]

        statements = []
        for stmt in stmts.body:
            if m.matches(stmt, is_chain_stmt):
                chain = ChainParser.from_parent(self)
                stmt.visit(chain)
                statements.append(chain.node)
            else:
                statements.append(self.make_verbatim_stmt(stmt))
        return statements

    def make_verbatim_stmt(self, cst_node):
        return self.make_node(VerbatimStatement, cst_node)


class ChainParser(ParserBase):
    """visits a chain statement, stores resulting node in self.node"""

    node: ChainStatement

    @property
    def _chain(self) -> t.List[ChainStep]:
        """convenience to get current chain"""
        return self.node.chain

    def _append(self, step: ChainStep) -> None:
        """convenience to append to chain"""
        self._chain.append(step)

    def on_visit(self, cst_node):
        if not hasattr(self, "node"):
            # we only reach this on the very first visit
            if not isinstance(cst_node, cst.BaseSmallStatement):
                warn(
                    "used ChainParser to visit a non-statement: "
                    f"{self.code_for(cst_node)}"
                )
            self.node = self.make_node(ChainStatement, cst_node, chain=[])

        # special case for top-level comparison-exprs to make sure we don't
        # recurse into them
        if m.matches(cst_node, is_bool_expr):
            bool_expr_parser = BoolExprParser.from_parent(self)
            cst_node.visit(bool_expr_parser)
            for step in bool_expr_parser.steps:
                self._append(step)
            # don't recurse down rest of tree
            return False

        return not (
            # skip all arguments so we don't recurse into nested function calls
            # and subscripts
            m.matches(cst_node, is_argument)
            # also skip the lhs of the = for assignments
            or isinstance(cst_node, cst.AssignTarget)
        )

    def leave_Subscript(self, cst_node: cst.Subscript):
        # special case: use separate parser for subscripts since they require
        # different logic than function calls
        subs_parser = SubscriptParser.from_parent(self)
        cst_node.visit(subs_parser)
        self._append(subs_parser.step)

    def fallback_call(self, cst_node):
        """
        called whenever we don't know how to parse a call. that could be
        when we don't handle the function, or if the function has weird
        arguments that we can't parse.
        """
        func = cst_node.func.attr.value
        step = self.make_call_node(PassThroughCall, cst_node, func=func)
        self._append(step)

    # for things in a chain, we append to chain on _leaving_ a node because of
    # the way chains are nested in the CST. for a chain like df.a().b(), the
    # parse order is b(), then a(), then df. if we append on leaving,
    # then we get the right chain order of df, a(), b().

    # gets the df out of:
    # df.f()
    # df['hello']
    @m.leave(m.Name())
    def make_chain_start(self, cst_node):
        # we should only do this for the first name in an chain, since for
        # `df.assign` both `df` and `assign` are Names
        if len(self._chain) == 0:
            step = self.make_node(StartOfChain, cst_node)
            self._append(step)

    @m.leave(is_attribute_call)
    def make_pass_through_call(self, cst_node):
        if is_parsed_call(cst_node):
            return

        # special case: make an AggCall if the last call was a GroupBy and
        # we're using an agg function
        if len(self._chain) > 1:
            last = self._chain[-1]
            if isinstance(last, GroupByCall):
                func_name: str = cst_node.func.attr.value
                if func_name in AggCall.agg_funcs:
                    node = self.make_call_node(AggCall, cst_node)
                    self._append(node)
                    return
        # TODO: handle transforming functions like `.transform`

        self.fallback_call(cst_node)

    @m.leave(matches(HeadCall) | matches(TailCall))
    def make_head_or_tail(self, cst_node):
        name = fn_name(cst_node)
        node = self.make_call_node(
            HeadCall if name == "head" else TailCall, cst_node
        )
        self._append(node)

    @m.leave(matches(GetCall))
    def make_get_call(self, cst_node):
        labels_expr = self.get_arg_code(cst_node.args, 0, "key")

        node = self.make_call_node(
            GetCall,
            cst_node,
            labels_expr=labels_expr,
        )
        self._append(node)

    @m.leave(matches(SortValuesCall))
    def make_sort_values_call(self, cst_node):
        label_expr = self.get_arg_code(cst_node.args, 0, "by")
        axis_arg = self.get_arg_code(cst_node.args, 1, "axis")
        axis = make_axis(axis_arg) if axis_arg is not None else "index"

        node = self.make_call_node(
            SortValuesCall, cst_node, label_expr=label_expr, axis=axis
        )
        self._append(node)

    @m.leave(matches(DropCall))
    def make_drop_call(self, cst_node):
        labels = self.get_arg_code(cst_node.args, 0, "labels")
        axis_arg = self.get_arg_code(cst_node.args, 1, "axis")
        row_expr = self.get_arg_code(cst_node.args, 2, "index")
        col_expr = self.get_arg_code(cst_node.args, 3, "columns")
        axis = make_axis(axis_arg) if axis_arg is not None else "index"

        if labels is not None:
            if axis == "columns":
                col_expr = labels
            else:
                row_expr = labels

        node = self.make_call_node(
            DropCall, cst_node, col_expr=col_expr, row_expr=row_expr
        )
        self._append(node)

    @m.leave(matches(RenameCall))
    def make_rename_call(self, cst_node):
        mapper = get_arg(cst_node.args, 0, "mapper")
        index = get_arg(cst_node.args, 1, "index")
        columns = get_arg(cst_node.args, 2, "columns")
        axis_arg = get_arg(cst_node.args, 3, "axis")
        axis = "index"  # default

        if index is not None:
            mapper = index
            axis = "index"
        elif columns is not None:
            mapper = columns
            axis = "columns"
        elif axis_arg is not None:
            axis = make_axis(self.code_for(axis_arg.value))

        if mapper is None:
            self.fallback_call(cst_node)
            return

        node = self.make_call_node(
            RenameCall,
            cst_node,
            mapping_expr=self.code_for(mapper.value),
            axis=axis,
        )
        self._append(node)

    @m.leave(matches(ApplyCall))
    def make_apply(self, cst_node):
        # axis only available for dataframes...for series, arg 1 is some other
        # arg that we don't care about so we should be careful here
        axis_arg = self.get_arg_code(cst_node.args, 1, "axis")
        axis = make_axis(axis_arg) if axis_arg is not None else "index"

        node = self.make_call_node(ApplyCall, cst_node, axis=axis)
        self._append(node)

    @m.leave(matches(AssignCall))
    def make_assign(self, cst_node: cst.Call):
        # each kwarg is a new column
        new_col_labels = [
            arg.keyword.value
            for arg in cst_node.args
            if arg.keyword is not None
        ]

        node = self.make_call_node(
            AssignCall, cst_node, new_col_labels=new_col_labels
        )
        self._append(node)

    @m.leave(matches(GroupByCall))
    def make_groupby(self, cst_node):
        axis_arg = self.get_arg_code(cst_node.args, 1, "axis")
        axis = make_axis(axis_arg) if axis_arg is not None else "index"

        node = self.make_call_node(GroupByCall, cst_node, axis=axis)
        self._append(node)

    @m.leave(matches(ResetIndexCall))
    def make_reset_index(self, cst_node):
        level_expr = self.get_arg_code(cst_node.args, 0, "level")
        drop_expr = self.get_arg_code(cst_node.args, 1, "drop")

        node = self.make_call_node(
            ResetIndexCall,
            cst_node,
            level_expr=level_expr,
            drop_expr=drop_expr,
        )
        self._append(node)

    @m.leave(matches(SetIndexCall))
    def make_set_index(self, cst_node):
        keys_expr = self.get_arg_code(cst_node.args, 0, "keys")
        append_expr = self.get_arg_code(cst_node.args, 2, "append")

        node = self.make_call_node(
            SetIndexCall,
            cst_node,
            keys_expr=keys_expr,
            append_expr=append_expr,
        )
        self._append(node)

    @m.leave(matches(UnstackCall))
    def make_unstack(self, cst_node):
        level_expr = self.get_arg_code(cst_node.args, 0, "level")

        node = self.make_call_node(
            UnstackCall,
            cst_node,
            level_expr=level_expr,
        )
        self._append(node)

    @m.leave(matches(StackCall))
    def make_stack(self, cst_node):
        level_expr = self.get_arg_code(cst_node.args, 0, "level")

        node = self.make_call_node(
            StackCall,
            cst_node,
            level_expr=level_expr,
        )
        self._append(node)

    @m.leave(matches(PivotCall))
    def make_pivot(self, cst_node: cst.Call):
        index_expr = self.get_arg_code(cst_node.args, 0, "index")
        columns_expr = self.get_arg_code(cst_node.args, 1, "columns")
        values_expr = self.get_arg_code(cst_node.args, 2, "values")

        node = self.make_call_node(
            PivotCall,
            cst_node,
            index_expr=index_expr,
            columns_expr=columns_expr,
            values_expr=values_expr,
        )
        self._append(node)

    @m.leave(matches(PivotTableCall))
    def make_pivot_table(self, cst_node: cst.Call):
        values_expr = self.get_arg_code(cst_node.args, 0, "values")
        index_expr = self.get_arg_code(cst_node.args, 1, "index")
        columns_expr = self.get_arg_code(cst_node.args, 2, "columns")
        aggfunc_expr = self.get_arg_code(cst_node.args, 3, "aggfunc")

        node = self.make_call_node(
            PivotTableCall,
            cst_node,
            index_expr=index_expr,
            columns_expr=columns_expr,
            values_expr=values_expr,
            aggfunc_expr=aggfunc_expr,
        )
        self._append(node)

    @m.leave(matches(MeltCall))
    def make_melt(self, cst_node: cst.Call):
        id_vars_expr = self.get_arg_code(cst_node.args, 0, "id_vars")
        value_vars_expr = self.get_arg_code(cst_node.args, 1, "value_vars")
        var_name_expr = self.get_arg_code(cst_node.args, 2, "var_name")
        value_name_expr = self.get_arg_code(cst_node.args, 3, "value_name")
        col_level_expr = self.get_arg_code(cst_node.args, 4, "col_level")
        ignore_index_expr = self.get_arg_code(cst_node.args, 5, "ignore_index")

        node = self.make_call_node(
            MeltCall,
            cst_node,
            id_vars_expr=id_vars_expr,
            value_vars_expr=value_vars_expr,
            var_name_expr=var_name_expr,
            value_name_expr=value_name_expr,
            col_level_expr=col_level_expr,
            ignore_index_expr=ignore_index_expr,
        )
        self._append(node)

    @m.leave(matches(MergeCall))
    def make_merge(self, cst_node: cst.Call):
        right_expr = self.get_arg_code(cst_node.args, 0, "right")
        how_expr = self.get_arg_code(cst_node.args, 1, "how")
        on_expr = self.get_arg_code(cst_node.args, 2, "on")
        left_on_expr = self.get_arg_code(cst_node.args, 3, "left_on")
        right_on_expr = self.get_arg_code(cst_node.args, 4, "right_on")
        left_index_expr = self.get_arg_code(cst_node.args, 5, "left_index")
        right_index_expr = self.get_arg_code(cst_node.args, 6, "right_index")
        sort_expr = self.get_arg_code(cst_node.args, 7, "sort")
        suffixes_expr = self.get_arg_code(cst_node.args, 8, "suffixes")

        node = self.make_call_node(
            MergeCall,
            cst_node,
            right_expr=right_expr,
            how_expr=how_expr,
            on_expr=on_expr,
            left_on_expr=left_on_expr,
            right_on_expr=right_on_expr,
            left_index_expr=left_index_expr,
            right_index_expr=right_index_expr,
            sort_expr=sort_expr,
            suffixes_expr=suffixes_expr,
        )
        self._append(node)

    @m.leave(matches(JoinCall))
    def make_join(self, cst_node: cst.Call):
        other_expr = self.get_arg_code(cst_node.args, 0, "other")
        on_expr = self.get_arg_code(cst_node.args, 1, "on")
        how_expr = self.get_arg_code(cst_node.args, 2, "how")
        sort_expr = self.get_arg_code(cst_node.args, 5, "sort")

        node = self.make_call_node(
            JoinCall,
            cst_node,
            other_expr=other_expr,
            on_expr=on_expr,
            how_expr=how_expr,
            sort_expr=sort_expr,
        )
        self._append(node)

    def make_call_node(self, cls: t.Type[T], cst_node: cst.Call, **kwargs) -> T:
        """
        for calls in chain like df.apply(), the location of the call is the
        dot + everything after
        """
        func = t.cast(cst.Attribute, cst_node.func)
        dot = self.location(func.dot)
        entire_expr = self.location(cst_node)
        location = CodeRange(dot.start, entire_expr.end)

        return self.make_node(cls, cst_node, location=location, **kwargs)

    def get_arg_code(
        self,
        args: t.Sequence[cst.Arg],
        position: int,
        keyword: t.Optional[str] = None,
    ) -> t.Optional[RawCode]:
        arg = get_arg(args, position, keyword)
        return arg if arg is None else self.code_for(arg.value)


class BoolExprParser(ParserBase):
    """
    special parser for boolean expressions, stores resulting steps in
    self.steps

    the key idea is to parse only one level deep to avoid making super
    complicated diagrams. so expressions like this one:

    ((df.get('Count') > 13000) &
     (df.get('Count') < 15000) &
     (df.get('Sex') == 'M'))

    get parsed into 3 steps, not 5. if users want to look at the inner
    comparisons, they should visualize those separately

    the big assumption is that boolean expressions can only appear at the start
    of a function call chain, not in the middle.
    """

    steps: t.List[ChainStep]

    def on_visit(self, cst_node):
        self.steps = []

        if m.matches(cst_node, m.BinaryOperation()):
            self.parse_binary_op(cst_node)
        elif m.matches(cst_node, m.Comparison()):
            self.parse_comparison(cst_node)
        else:
            warn(
                "used BoolExprParser to visit a non-boolean expr: "
                f"{self.code_for(cst_node)}"
            )

        # don't use libcst to recurse
        return False

    # boolean exprs are guaranteed to be at the start of a chain since we
    # don't recurse into them in ChainParser, so we need to make a
    # StartOfChain node even though we'll use the lhs attribute to make
    # diagrams instead.

    # matches '|' and '&' operators
    def parse_binary_op(self, cst_node: cst.BinaryOperation):
        if not m.matches(cst_node.left, m.BinaryOperation()):
            self.steps.append(self.make_node(StartOfChain, cst_node.left))
        else:
            self.parse_binary_op(t.cast(cst.BinaryOperation, cst_node.left))

        self.steps.append(self.make_node(BoolExprStep, cst_node))

    # comparisons don't need to recurse
    def parse_comparison(self, cst_node: cst.Comparison):
        self.steps.append(self.make_node(StartOfChain, cst_node.left))
        start = self.location(cst_node.left)

        # each comp is just the ' > 13000' part of the expr, so we need to make
        # sure the lhs is also in code for execution
        current_code = self.code_for(cst_node.left)
        for comp in cst_node.comparisons:
            end = self.location(comp)
            current_code += self.code_for(comp)
            self.steps.append(
                self.make_node(
                    BoolExprStep, comp, code=current_code, location=start | end
                )
            )


class SubscriptParser(ParserBase):
    """
    special parser for subscript expressions, stores resulting node in
    self.step
    """

    step: Subscript

    def on_visit(self, node):
        if not isinstance(node, cst.Subscript):
            warn(
                "used SubscriptParser to visit a non-subscript: "
                f"{self.code_for(node)}"
            )
            return False

        slicer: t.Optional[str] = None
        if m.matches(node, is_loc_iloc):
            slicer = t.cast(cst.Attribute, node.value).attr.value

        # from dogs["breed"], get location of ["breed"]
        location = self.location(node.lbracket) | self.location(node.rbracket)

        # if we have a slicer, then we want the '.loc' too
        if slicer is not None:
            location = location | self.location(
                t.cast(cst.Attribute, node.value).dot
            )

        n_slices = len(node.slice)
        slice1 = (
            self.parse_slice(node.slice[0].slice) if n_slices >= 1 else None
        )
        slice2 = (
            self.parse_slice(node.slice[1].slice) if n_slices >= 2 else None
        )
        if n_slices > 2:
            warn(
                f"weird: parsed subscript with {n_slices} slices @\n"
                f"{self.code_for(node)}"
            )

        self.step = self.make_node(
            Subscript,
            node,
            location=location,
            slicer=slicer,
            slice1=slice1,
            slice2=slice2,
        )

        # We do everything within on_visit
        return False

    def parse_slice(self, node: cst.BaseSlice) -> t.Optional[SubscriptEl]:
        if m.matches(node, m.Slice()):
            return self.make_node(SubsSlice, node)
        elif m.matches(node, m.Index(value=is_boolean_slice)):
            node = t.cast(cst.Index, node)
            extractor = ExtractLiteralIndexes.from_parent(self)
            node.value.visit(extractor)
            return self.make_node(
                SubsComparison,
                node,
                label_exprs=extractor.labels,
            )
        elif m.matches(node, m.Index(value=~is_boolean_slice)):
            # the fallback case: just eval the slice expression
            return self.make_node(SubsEval, node, expr=self.code_for(node))
        else:
            warn(
                f"weird slice type, defaulting to empty slice:\n"
                f"{self.code_for(node)}"
            )
            return self.make_node(SubsSlice, node)


class ExtractLiteralIndexes(ParserBase):
    """
    special parser to get all simple index values out of a subscript index.
    pandas has expressions that look like this:

        df2[df2["E"].isin(["two", "four"])]

    where we want to get the "E" label out of the inner subscript.

    so we use this parser on Expr nodes to get:

    df[0]                                 -> ["0"]
    df['hello']                           -> ["'hello'"]
    (df[col] > 0) & (df['sam'] == 2)      -> ["col", "'sam'"]
    (df[df['nested']]) & (df['sam'] == 2) -> ["'sam'"]

    but the current implementation doesn't try to distinguish iloc from loc:

    df.iloc[0]                            -> ["0"]
    """

    labels: t.List[RawCode]
    _depth: int

    @property
    def _is_nested(self):
        return self._depth > 1

    def on_visit(self, cst_node):
        if not hasattr(self, "labels"):
            if not isinstance(cst_node, cst.BaseExpression):
                warn(
                    "used ExtractLiteralIndexes to visit a non-expression: "
                    f"{self.code_for(cst_node)}"
                )
            self.labels = []
            self._depth = 0
        return super().on_visit(cst_node)

    def visit_Subscript(self, cst_node):
        self._depth += 1

    def leave_Subscript(self, cst_node):
        self._depth -= 1

    def visit_Call(self, cst_node):
        self._depth += 1

    def leave_Call(self, cst_node):
        self._depth -= 1

    @m.visit(m.Index())
    def record_index(self, cst_node: cst.Index):
        if self._is_nested:
            return

        val = cst_node.value
        if isinstance(val, (cst.BaseString, cst.BaseNumber, cst.Name)):
            self.labels.append(self.code_for(val))

    @m.call_if_inside(matches(GetCall))
    @m.leave(m.Arg())
    def record_get_arg(self, cst_node: cst.Arg):
        if self._is_nested:
            return

        val = cst_node.value
        if isinstance(val, (cst.BaseString, cst.BaseNumber, cst.Name)):
            self.labels.append(self.code_for(val))


whitespace = (
    m.Comment()
    | m.EmptyLine()
    | m.Newline()
    | m.ParenthesizedWhitespace()
    | m.SimpleWhitespace()
    | m.TrailingWhitespace()
    | m.BaseParenthesizableWhitespace()
)


# For debugging; it just logs the nodes it visits
class LoggingVisitor(m.MatcherDecoratableVisitor):
    METADATA_DEPENDENCIES = (cstm.PositionProvider,)

    cst_root: t.Optional[cst.Module]
    depth: int

    def __init__(self):
        self.depth = 0
        self.cst_root = None
        super().__init__()

    def on_visit(self, node):
        if m.matches(node, whitespace):
            return False
        if self.cst_root is None:
            self.cst_root = t.cast(cst.Module, node)
            return True

        self.log(node)
        self.depth += 1
        return True

    def on_leave(self, node):
        if m.matches(node, whitespace):
            return
        self.depth -= 1

    def log(self, node):
        assert self.cst_root is not None
        name = node.__class__.__name__
        code = self.cst_root.code_for_node(node)
        # meta = self.get_metadata(cstm.PositionProvider, node)
        # start = meta.start
        # end = meta.end

        spaces = "  " * self.depth
        print(f'{spaces + name + ":": <40} ({code})')
        # print(f'{spaces + name + ":": <20} ({start.line}, {start.column}) '
        #       f'-> ({end.line}, {end.column})')


def test_logger(code):
    print(code)
    print("\n-----\n")
    tree = cst.parse_module(code)
    with_meta = cstm.MetadataWrapper(tree)
    sam = LoggingVisitor()
    _ = with_meta.visit(sam)
    return


def test_parser(code):
    print(code)
    print("\n-----\n")
    tree = cst.parse_module(code)
    with_meta = cstm.MetadataWrapper(tree)
    positions = with_meta.resolve(cstm.PositionProvider)
    sam = PandasParser(tree, positions)
    _ = with_meta.visit(sam)
    return sam.root



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import sys

# PEP 585
if sys.version_info < (3, 9):
    from typing import Iterable, Sequence
else:
    from collections.abc import Iterable, Sequence

from libcst._types import CSTNodeT_co


class FlattenSentinel(Sequence[CSTNodeT_co]):
    """
    A :class:`FlattenSentinel` may be returned by a :meth:`CSTTransformer.on_leave`
    method when one wants to replace a node with multiple nodes. The replaced
    node must be contained in a `Sequence` attribute such as
    :attr:`~libcst.Module.body`.  This is generally the case for
    :class:`~libcst.BaseStatement` and :class:`~libcst.BaseSmallStatement`.
    For example to insert a print before every return::

        def leave_Return(
            self, original_node: cst.Return, updated_node: cst.Return
        ) -> Union[cst.Return, cst.RemovalSentinel, cst.FlattenSentinel[cst.BaseSmallStatement]]:
            log_stmt = cst.Expr(cst.parse_expression("print('returning')"))
            return cst.FlattenSentinel([log_stmt, updated_node])

    Returning an empty :class:`FlattenSentinel` is equivalent to returning
    :attr:`cst.RemovalSentinel.REMOVE` and is subject to its requirements.
    """

    nodes: Sequence[CSTNodeT_co]

    def __init__(self, nodes: Iterable[CSTNodeT_co]) -> None:
        self.nodes = tuple(nodes)

    def __getitem__(self, idx: int) -> CSTNodeT_co:
        return self.nodes[idx]

    def __len__(self) -> int:
        return len(self.nodes)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# pyre-strict
import argparse
from typing import Callable, Optional, Sequence, Set, Tuple, Union

import libcst as cst
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand
from libcst.codemod.visitors import AddImportsVisitor, RemoveImportsVisitor
from libcst.helpers import get_full_name_for_node
from libcst.metadata import QualifiedNameProvider


def leave_import_decorator(
    method: Callable[..., Union[cst.Import, cst.ImportFrom]]
) -> Callable[..., Union[cst.Import, cst.ImportFrom]]:
    # We want to record any 'as name' that is relevant but only after we leave the corresponding Import/ImportFrom node since
    # we don't want the 'as name' to interfere with children 'Name' and 'Attribute' nodes.
    def wrapper(
        self: "RenameCommand",
        original_node: Union[cst.Import, cst.ImportFrom],
        updated_node: Union[cst.Import, cst.ImportFrom],
    ) -> Union[cst.Import, cst.ImportFrom]:
        updated_node = method(self, original_node, updated_node)
        if original_node != updated_node:
            self.record_asname(original_node)
        return updated_node

    return wrapper


class RenameCommand(VisitorBasedCodemodCommand):
    """
    Rename all instances of a local or imported object.
    """

    DESCRIPTION: str = "Rename all instances of a local or imported object."

    METADATA_DEPENDENCIES = (QualifiedNameProvider,)

    @staticmethod
    def add_args(parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--old_name",
            dest="old_name",
            required=True,
            help="Full dotted name of object to rename. Eg: `foo.bar.baz`",
        )

        parser.add_argument(
            "--new_name",
            dest="new_name",
            required=True,
            help=(
                "Full dotted name of replacement object. You may provide a single-colon-delimited name to specify how you want the new import to be structured."
                + "\nEg: `foo:bar.baz` will be translated to `from foo import bar`."
                + "\nIf no ':' character is provided, the import statement will default to `from foo.bar import baz` for a `new_name` value of `foo.bar.baz`"
                + " or simply replace the old import on the spot if the old import is an exact match."
            ),
        )

    def __init__(self, context: CodemodContext, old_name: str, new_name: str) -> None:
        super().__init__(context)

        new_module, has_colon, new_mod_or_obj = new_name.rpartition(":")
        # Exit early if improperly formatted args.
        if ":" in new_module:
            raise ValueError("Error: `new_name` should contain at most one colon.")
        if ":" in old_name:
            raise ValueError("Error: `old_name` should not contain any colons.")

        if not has_colon or not new_module:
            new_module, _, new_mod_or_obj = new_name.rpartition(".")

        self.new_name: str = new_name.replace(":", ".").strip(".")
        self.new_module: str = new_module.replace(":", ".").strip(".")
        self.new_mod_or_obj: str = new_mod_or_obj

        # If `new_name` contains a single colon at the end, then we assume the user wants the import
        # to be structured as 'import new_name'. So both self.new_mod_or_obj and self.old_mod_or_obj
        # will be empty in this case.
        if not self.new_mod_or_obj:
            old_module = old_name
            old_mod_or_obj = ""
        else:
            old_module, _, old_mod_or_obj = old_name.rpartition(".")

        self.old_name: str = old_name
        self.old_module: str = old_module
        self.old_mod_or_obj: str = old_mod_or_obj

        self.as_name: Optional[Tuple[str, str]] = None

        # A set of nodes that have been renamed to help with the cleanup of now potentially unused
        # imports, during import cleanup in `leave_Module`.
        self.scheduled_removals: Set[cst.CSTNode] = set()
        # If an import has been renamed while inside an `Import` or `ImportFrom` node, we want to flag
        # this so that we do not end up with two of the same import.
        self.bypass_import = False

    def visit_Import(self, node: cst.Import) -> None:
        for import_alias in node.names:
            alias_name = get_full_name_for_node(import_alias.name)
            if alias_name is not None:
                if alias_name == self.old_name or alias_name.startswith(
                    self.old_name + "."
                ):
                    # If the import statement is exactly equivalent to the old name, or we are renaming a top-level module of the import,
                    # it will be taken care of in `leave_Name` or `leave_Attribute` when visiting the Name and Attribute children of this Import.
                    self.bypass_import = True

    @leave_import_decorator
    def leave_Import(
        self, original_node: cst.Import, updated_node: cst.Import
    ) -> cst.Import:
        new_names = []
        for import_alias in updated_node.names:
            import_alias_name = import_alias.name
            import_alias_full_name = get_full_name_for_node(import_alias_name)
            if import_alias_full_name is None:
                raise Exception("Could not parse full name for ImportAlias.name node.")

            if isinstance(import_alias_name, cst.Name) and self.old_name.startswith(
                import_alias_full_name + "."
            ):
                # Might, be in use elsewhere in the code, so schedule a potential removal, and add another alias.
                new_names.append(import_alias)
                self.scheduled_removals.add(original_node)
                new_names.append(
                    cst.ImportAlias(
                        name=cst.Name(
                            value=self.gen_replacement_module(import_alias_full_name)
                        )
                    )
                )
                self.bypass_import = True
            elif isinstance(
                import_alias_name, cst.Attribute
            ) and self.old_name.startswith(import_alias_full_name + "."):
                # Same idea as above.
                new_names.append(import_alias)
                self.scheduled_removals.add(original_node)
                new_name_node: Union[
                    cst.Attribute, cst.Name
                ] = self.gen_name_or_attr_node(
                    self.gen_replacement_module(import_alias_full_name)
                )
                new_names.append(cst.ImportAlias(name=new_name_node))
                self.bypass_import = True
            else:
                new_names.append(import_alias)

        return updated_node.with_changes(names=new_names)

    def visit_ImportFrom(self, node: cst.ImportFrom) -> None:
        module = node.module
        if module is None:
            return
        imported_module_name = get_full_name_for_node(module)
        if imported_module_name is None:
            return
        if imported_module_name == self.old_name or imported_module_name.startswith(
            self.old_name + "."
        ):
            # If the imported module is exactly equivalent to the old name or we are renaming a parent module of the current module,
            # it will be taken care of in `leave_Name` or `leave_Attribute` when visiting the children of this ImportFrom.
            self.bypass_import = True

    @leave_import_decorator
    def leave_ImportFrom(
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> cst.ImportFrom:
        module = updated_node.module
        if module is None:
            return updated_node
        imported_module_name = get_full_name_for_node(module)
        names = original_node.names

        if imported_module_name is None or not isinstance(names, Sequence):
            return updated_node

        else:
            new_names = []
            for import_alias in names:
                alias_name = get_full_name_for_node(import_alias.name)
                if alias_name is not None:
                    qual_name = f"{imported_module_name}.{alias_name}"
                    if self.old_name == qual_name:

                        replacement_module = self.gen_replacement_module(
                            imported_module_name
                        )
                        replacement_obj = self.gen_replacement(alias_name)
                        if not replacement_obj:
                            # The user has requested an `import` statement rather than an `from ... import`.
                            # This will be taken care of in `leave_Module`, in the meantime, schedule for potential removal.
                            new_names.append(import_alias)
                            self.scheduled_removals.add(original_node)
                            continue

                        new_import_alias_name: Union[
                            cst.Attribute, cst.Name
                        ] = self.gen_name_or_attr_node(replacement_obj)
                        # Rename on the spot only if this is the only imported name under the module.
                        if len(names) == 1:
                            self.bypass_import = True
                            return updated_node.with_changes(
                                module=cst.parse_expression(replacement_module),
                                names=(cst.ImportAlias(name=new_import_alias_name),),
                            )
                        # Or if the module name is to stay the same.
                        elif replacement_module == imported_module_name:
                            self.bypass_import = True
                            new_names.append(
                                cst.ImportAlias(name=new_import_alias_name)
                            )
                    else:
                        if self.old_name.startswith(qual_name + "."):
                            # This import might be in use elsewhere in the code, so schedule a potential removal.
                            self.scheduled_removals.add(original_node)
                        new_names.append(import_alias)

            return updated_node.with_changes(names=new_names)
        return updated_node

    def leave_Name(
        self, original_node: cst.Name, updated_node: cst.Name
    ) -> Union[cst.Attribute, cst.Name]:
        full_name_for_node: str = original_node.value
        full_replacement_name = self.gen_replacement(full_name_for_node)

        # If a node has no associated QualifiedName, we are still inside an import statement.
        inside_import_statement: bool = not self.get_metadata(
            QualifiedNameProvider, original_node, set()
        )
        if QualifiedNameProvider.has_name(self, original_node, self.old_name) or (
            inside_import_statement and full_replacement_name == self.new_name
        ):
            if not full_replacement_name:
                full_replacement_name = self.new_name
            if not inside_import_statement:
                self.scheduled_removals.add(original_node)
            return self.gen_name_or_attr_node(full_replacement_name)

        return updated_node

    def leave_Attribute(
        self, original_node: cst.Attribute, updated_node: cst.Attribute
    ) -> Union[cst.Name, cst.Attribute]:
        full_name_for_node = get_full_name_for_node(original_node)
        if full_name_for_node is None:
            raise Exception("Could not parse full name for Attribute node.")
        full_replacement_name = self.gen_replacement(full_name_for_node)

        # If a node has no associated QualifiedName, we are still inside an import statement.
        inside_import_statement: bool = not self.get_metadata(
            QualifiedNameProvider, original_node, set()
        )
        if (
            QualifiedNameProvider.has_name(
                self,
                original_node,
                self.old_name,
            )
            or (inside_import_statement and full_replacement_name == self.new_name)
        ):
            new_value, new_attr = self.new_module, self.new_mod_or_obj
            if not inside_import_statement:
                self.scheduled_removals.add(original_node.value)
            if full_replacement_name == self.new_name:
                return updated_node.with_changes(
                    value=cst.parse_expression(new_value),
                    attr=cst.Name(value=new_attr.rstrip(".")),
                )

            return self.gen_name_or_attr_node(new_attr)

        return updated_node

    def leave_Module(
        self, original_node: cst.Module, updated_node: cst.Module
    ) -> cst.Module:
        for removal_node in self.scheduled_removals:
            RemoveImportsVisitor.remove_unused_import_by_node(
                self.context, removal_node
            )
        # If bypass_import is False, we know that no import statements were directly renamed, and the fact
        # that we have any `self.scheduled_removals` tells us we encountered a matching `old_name` in the code.
        if not self.bypass_import and self.scheduled_removals:
            if self.new_module:
                new_obj: Optional[str] = (
                    self.new_mod_or_obj.split(".")[0] if self.new_mod_or_obj else None
                )
                AddImportsVisitor.add_needed_import(
                    self.context, module=self.new_module, obj=new_obj
                )
        return updated_node

    def gen_replacement(self, original_name: str) -> str:
        module_as_name = self.as_name
        if module_as_name is not None:
            if original_name == module_as_name[0]:
                original_name = module_as_name[1]
            elif original_name.startswith(module_as_name[0] + "."):
                original_name = original_name.replace(
                    module_as_name[0] + ".", module_as_name[1] + ".", 1
                )

        if original_name == self.old_mod_or_obj:
            return self.new_mod_or_obj
        elif original_name == ".".join([self.old_module, self.old_mod_or_obj]):
            return self.new_name
        elif original_name.endswith("." + self.old_mod_or_obj):
            return self.new_mod_or_obj
        else:
            return self.gen_replacement_module(original_name)

    def gen_replacement_module(self, original_module: str) -> str:
        return self.new_module if original_module == self.old_module else ""

    def gen_name_or_attr_node(
        self, dotted_expression: str
    ) -> Union[cst.Attribute, cst.Name]:
        name_or_attr_node: cst.BaseExpression = cst.parse_expression(dotted_expression)
        if not isinstance(name_or_attr_node, (cst.Name, cst.Attribute)):
            raise Exception(
                "`parse_expression()` on dotted path returned non-Attribute-or-Name."
            )
        return name_or_attr_node

    def record_asname(self, original_node: Union[cst.Import, cst.ImportFrom]) -> None:
        # Record the import's `as` name if it has one, and set the attribute mapping.
        names = original_node.names
        if not isinstance(names, Sequence):
            return
        for import_alias in names:
            alias_name = get_full_name_for_node(import_alias.name)
            if isinstance(original_node, cst.ImportFrom):
                module = original_node.module
                if module is None:
                    return
                module_name = get_full_name_for_node(module)
                if module_name is None:
                    return
                qual_name = f"{module_name}.{alias_name}"
            else:
                qual_name = alias_name
            if qual_name is not None and alias_name is not None:
                if qual_name == self.old_name or self.old_name.startswith(
                    qual_name + "."
                ):
                    as_name_optional = import_alias.asname
                    as_name_node = (
                        as_name_optional.name if as_name_optional is not None else None
                    )
                    if as_name_node is not None and isinstance(
                        as_name_node, (cst.Name, cst.Attribute)
                    ):
                        full_as_name = get_full_name_for_node(as_name_node)
                        if full_as_name is not None:
                            self.as_name = (full_as_name, alias_name)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from libcst.codemod.visitors._add_imports import AddImportsVisitor
from libcst.codemod.visitors._apply_type_annotations import ApplyTypeAnnotationsVisitor
from libcst.codemod.visitors._gather_comments import GatherCommentsVisitor
from libcst.codemod.visitors._gather_exports import GatherExportsVisitor
from libcst.codemod.visitors._gather_imports import GatherImportsVisitor
from libcst.codemod.visitors._gather_string_annotation_names import (
    GatherNamesFromStringAnnotationsVisitor,
)
from libcst.codemod.visitors._gather_unused_imports import GatherUnusedImportsVisitor
from libcst.codemod.visitors._remove_imports import RemoveImportsVisitor

__all__ = [
    "AddImportsVisitor",
    "ApplyTypeAnnotationsVisitor",
    "GatherCommentsVisitor",
    "GatherExportsVisitor",
    "GatherImportsVisitor",
    "GatherNamesFromStringAnnotationsVisitor",
    "GatherUnusedImportsVisitor",
    "RemoveImportsVisitor",
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import fields
from typing import List

from libcst.codegen.gather import imports, nodebases, nodeuses

generated_code: List[str] = []
generated_code.append("# Copyright (c) Facebook, Inc. and its affiliates.")
generated_code.append("#")
generated_code.append(
    "# This source code is licensed under the MIT license found in the"
)
generated_code.append("# LICENSE file in the root directory of this source tree.")
generated_code.append("")
generated_code.append("")
generated_code.append("# This file was generated by libcst.codegen.gen_matcher_classes")
generated_code.append("from typing import Optional, Union, TYPE_CHECKING")
generated_code.append("")
generated_code.append("from libcst._flatten_sentinel import FlattenSentinel")
generated_code.append("from libcst._maybe_sentinel import MaybeSentinel")
generated_code.append("from libcst._removal_sentinel import RemovalSentinel")
generated_code.append("from libcst._typed_visitor_base import mark_no_op")

# Import the types we use. These have to be type guarded since it would
# cause an import cycle otherwise.
generated_code.append("")
generated_code.append("")
generated_code.append("if TYPE_CHECKING:")
for module, objects in imports.items():
    generated_code.append(f"    from {module} import (  # noqa: F401")
    generated_code.append(f"        {', '.join(sorted(list(objects)))}")
    generated_code.append("    )")


# Generate the base visit_ methods
generated_code.append("")
generated_code.append("")
generated_code.append("class CSTTypedBaseFunctions:")
for node in sorted(nodebases.keys(), key=lambda node: node.__name__):
    name = node.__name__
    if name.startswith("Base"):
        continue

    generated_code.append("")
    generated_code.append("    @mark_no_op")
    generated_code.append(
        f'    def visit_{name}(self, node: "{name}") -> Optional[bool]:'
    )
    generated_code.append("        pass")
    for field in fields(node) or []:
        if field.name == "_metadata":
            continue
        generated_code.append("")
        generated_code.append("    @mark_no_op")
        generated_code.append(
            f'    def visit_{name}_{field.name}(self, node: "{name}") -> None:'
        )
        generated_code.append("        pass")
        generated_code.append("")
        generated_code.append("    @mark_no_op")
        generated_code.append(
            f'    def leave_{name}_{field.name}(self, node: "{name}") -> None:'
        )
        generated_code.append("        pass")

# Generate the visitor leave_ methods
generated_code.append("")
generated_code.append("")
generated_code.append("class CSTTypedVisitorFunctions(CSTTypedBaseFunctions):")
for node in sorted(nodebases.keys(), key=lambda node: node.__name__):
    name = node.__name__
    if name.startswith("Base"):
        continue

    generated_code.append("")
    generated_code.append("    @mark_no_op")
    generated_code.append(
        f'    def leave_{name}(self, original_node: "{name}") -> None:'
    )
    generated_code.append("        pass")

# Generate the transformer leave_ methods
generated_code.append("")
generated_code.append("")
generated_code.append("class CSTTypedTransformerFunctions(CSTTypedBaseFunctions):")
generated_code.append("    pass")
for node in sorted(nodebases.keys(), key=lambda node: node.__name__):
    name = node.__name__
    if name.startswith("Base"):
        continue
    generated_code.append("")
    generated_code.append("    @mark_no_op")
    valid_return_types: List[str] = [f'"{nodebases[node].__name__}"']
    node_uses = nodeuses[node]
    base_uses = nodeuses[nodebases[node]]
    if node_uses.maybe or base_uses.maybe:
        valid_return_types.append("MaybeSentinel")

    if node_uses.sequence or base_uses.sequence:
        valid_return_types.append(f'FlattenSentinel["{nodebases[node].__name__}"]')
        valid_return_types.append("RemovalSentinel")
    elif node_uses.optional or base_uses.optional:
        valid_return_types.append("RemovalSentinel")

    generated_code.append(
        f'    def leave_{name}(self, original_node: "{name}", updated_node: "{name}") -> Union[{", ".join(valid_return_types)}]:'
    )
    generated_code.append("        return updated_node")

if __name__ == "__main__":
    # Output the code
    print("\n".join(generated_code))



"""
node objects that parse.py creates
"""
from __future__ import annotations

import dataclasses
import json
from dataclasses import field
from typing import ClassVar, List, NewType, Optional, TypeVar, Union

from pandas_tutor.util import Axis, CodeRange, Slicer, pd_agg_funcs

T = TypeVar("T")

# Use a distinct type to distinguish between strings that can be eval'd
RawCode = NewType("RawCode", str)


class _ParseTreeEncoder(json.JSONEncoder):
    def default(self, obj):
        if dataclasses.is_dataclass(obj):
            return dataclasses.asdict(obj)
        return super().default(obj)


@dataclasses.dataclass
class Base:
    type_: str = field(init=False, repr=False)
    code: RawCode
    location: CodeRange

    def __post_init__(self):
        self.type_ = self.__class__.__name__

    def to_dict(self):
        return dataclasses.asdict(self)

    @classmethod
    def to_json(cls, items: Union[List[Base], Base]):
        return json.dumps(items, indent=2, cls=_ParseTreeEncoder)


@dataclasses.dataclass
class ParsedModule(Base):
    """root of parse tree"""

    statements: List[Statement]


@dataclasses.dataclass
class VerbatimStatement(Base):
    """node that we should just run, like imports"""

    pass


@dataclasses.dataclass
class ChainStatement(Base):
    """node that we have special parse rules for, like function chains"""

    # TODO: should we distinguish between Assign and Exprs?
    chain: List[ChainStep]


Statement = Union[VerbatimStatement, ChainStatement]


@dataclasses.dataclass
class ChainStep(Base):
    """represents a step that we can visualize, or an error"""

    pass


@dataclasses.dataclass
class StartOfChain(ChainStep):
    """the pd in pd.pivot_table, or df in df['Name']"""

    pass


##############################################################################
# Calls
#
# calls and subscripts have arguments that we need to eval. to handle this,
# we'll do the following:
#
# 1. when parsing, we save the raw code for arguments we'll eval later, and
#    mark those fields using the `evals_into` function (see
#    SortValuesCall.label_expr).
# 2. when executing, we eval each field and save the results into
#    run.EvalResult.args. so, we'll take evals_into('labels') and put it into
#    EvalResult.args['labels']
##############################################################################


@dataclasses.dataclass
class Call(ChainStep):
    """base class used for typing"""

    # used when parsing to match functions to Call subclasses
    fn_name: ClassVar


def evals_into(attr: str):
    """marks a field in a dataclass as one that should be eval'd"""
    return field(metadata=dict(evals_into=attr))


@dataclasses.dataclass
class GetCall(Call):
    """
    df.get('size')
    df.get(['size', 'breed'])
    df.get(['size', 'breed'], default=False)
    """

    fn_name = "get"

    # Expression that evaluates to labels
    labels_expr: RawCode = evals_into("labels")


@dataclasses.dataclass
class SortValuesCall(Call):
    """
    cols = ['size', 'breed']
    df.sort_values(cols)
    """

    fn_name = "sort_values"

    # Expression that evaluates to labels
    label_expr: RawCode = evals_into("labels")

    # technically the axis can be an expression too...but who does that??
    axis: Axis = "index"


@dataclasses.dataclass
class RenameCall(Call):
    """
    names = {'size': 'SIZE', 'food_cost': 'cost'}
    df.rename(names, axis=1)
    df.rename(axis=1, mapper=names)
    df.rename(columns=names)
    df.rename(index={'sam': 'smae'})
    """

    fn_name = "rename"

    # expression that results in a dict. can sometimes be a function, in which
    # case we need to just pass it through
    mapping_expr: RawCode = evals_into("mapping")

    axis: Axis = "index"


@dataclasses.dataclass
class HeadCall(Call):
    """
    df.head(5)
    df.head()
    df.head(-2)
    """

    fn_name = "head"


@dataclasses.dataclass
class TailCall(Call):
    """
    df.tail(5)
    df.tail()
    df.tail(-2)
    """

    fn_name = "tail"


@dataclasses.dataclass
class PassThroughCall(Call):
    """
    call that we don't know how to draw diagram for, so we should just run it
    and keep going
    """

    fn_name = "PASS_THROUGH"
    func: str


@dataclasses.dataclass
class GroupByCall(Call):
    """
    the labels for grouping are automatically saved into the groupby object,
    so we don't need to get them during parsing

    df.groupby('region')
    df.groupby(['region', 'id'])
    df.groupby(df['region'])
    df.groupby(lambda val: val // 10)
    """

    fn_name = "groupby"

    axis: Axis = "index"


@dataclasses.dataclass
class AggCall(Call):
    """
    catch-all for any function that happens after a groupby. note: some
    functions on groupby objects are transforms, not aggregations, e.g.
    .transform(), .apply(), .cumcount(), etc. and shouldn't be parsed into an
    AggCall

    g = df.groupby('region')
    g.agg('mean')
    g.mean()
    g.std()
    """

    # don't match the actual "agg()" call since we have a special case in the
    # parser for aggregations
    fn_name = "AGG"

    # exhaustive list of aggregation functions from GroupBy objects
    agg_funcs = ["agg", "aggregate", *pd_agg_funcs]

    @classmethod
    def from_passthrough_call(cls, call: PassThroughCall):
        return cls(code=call.code, location=call.location)


@dataclasses.dataclass
class ApplyCall(Call):
    """
    df['region'].apply(len)
    """

    fn_name = "apply"

    axis: Axis = "index"


@dataclasses.dataclass
class AssignCall(Call):
    """
    df.assign(test=2)
    df.assign(temp_f=df['temp_c'] * 9 / 5 + 32)
    """

    fn_name = "assign"

    new_col_labels: List[str]


@dataclasses.dataclass
class DropCall(Call):
    """
    dogs.drop(columns=['type', 'price'])
    dogs.drop(['Labrador Retriever', 'Beagle'])
    """

    fn_name = "drop"

    col_expr: Optional[RawCode] = evals_into("col_labels")
    row_expr: Optional[RawCode] = evals_into("row_labels")


@dataclasses.dataclass
class ResetIndexCall(Call):
    """
    dogs.reset_index(level=1)
    dogs.reset_index(level=[1, 2])
    dogs.reset_index(level=1, drop=True)
    """

    fn_name = "reset_index"

    level_expr: Optional[RawCode] = evals_into("level")
    drop_expr: Optional[RawCode] = evals_into("drop")


@dataclasses.dataclass
class SetIndexCall(Call):
    """
    dogs.set_index('breed', drop=False)
    dogs.set_index('breed', append=True)
    """

    fn_name = "set_index"

    keys_expr: Optional[RawCode] = evals_into("keys")
    # we don't need to track the drop arg
    append_expr: Optional[RawCode] = evals_into("append")


@dataclasses.dataclass
class UnstackCall(Call):
    """
    counts.unstack()
    counts.unstack(level=1)
    counts.unstack(level=[1, 2])
    counts.unstack(level=-1, fill_value=0)
    """

    fn_name = "unstack"

    level_expr: Optional[RawCode] = evals_into("level")


@dataclasses.dataclass
class StackCall(Call):
    """
    counts.stack()
    counts.stack(level=1)
    counts.stack(level=[1, 2])
    counts.stack(level=-1, dropna=True)
    """

    fn_name = "stack"

    level_expr: Optional[RawCode] = evals_into("level")


@dataclasses.dataclass
class PivotCall(Call):
    """
    df.pivot(index='foo', columns='bar', values='baz')
    """

    fn_name = "pivot"

    index_expr: Optional[RawCode] = evals_into("index")
    columns_expr: Optional[RawCode] = evals_into("columns")
    values_expr: Optional[RawCode] = evals_into("values")


@dataclasses.dataclass
class PivotTableCall(Call):
    """
    df.pivot_table(values='D', index=['A', 'B'],
                   columns=['C'], aggfunc=np.sum)
    df.pivot_table(values=['D', 'E'], index=['A', 'C'],
                   aggfunc={'D': np.mean,
                            'E': np.mean})
    """

    fn_name = "pivot_table"

    index_expr: Optional[RawCode] = evals_into("index")
    columns_expr: Optional[RawCode] = evals_into("columns")
    values_expr: Optional[RawCode] = evals_into("values")
    aggfunc_expr: Optional[RawCode] = evals_into("aggfunc")


@dataclasses.dataclass
class MeltCall(Call):
    """
    df.melt(id_vars=['A'], value_vars=['B'])
    df.melt(id_vars=['A'], value_vars=['B'],
            var_name='myVarname', value_name='myValname')
    """

    fn_name = "melt"

    id_vars_expr: Optional[RawCode] = evals_into("id_vars")
    value_vars_expr: Optional[RawCode] = evals_into("value_vars")
    var_name_expr: Optional[RawCode] = evals_into("var_name")
    value_name_expr: Optional[RawCode] = evals_into("value_name")
    col_level_expr: Optional[RawCode] = evals_into("col_level")
    ignore_index_expr: Optional[RawCode] = evals_into("ignore_index")


@dataclasses.dataclass
class MergeCall(Call):
    """
    baby.merge(nyt, left_on='Name', right_on='nyt_name', how='left')
    """

    fn_name = "merge"

    right_expr: Optional[RawCode] = evals_into("right")
    how_expr: Optional[RawCode] = evals_into("how")
    on_expr: Optional[RawCode] = evals_into("on")
    left_on_expr: Optional[RawCode] = evals_into("left_on")
    right_on_expr: Optional[RawCode] = evals_into("right_on")
    left_index_expr: Optional[RawCode] = evals_into("left_index")
    right_index_expr: Optional[RawCode] = evals_into("right_index")
    sort_expr: Optional[RawCode] = evals_into("sort")
    suffixes_expr: Optional[RawCode] = evals_into("suffixes")


@dataclasses.dataclass
class JoinCall(Call):
    """
    baby.join(nyt, on='Name', how='inner', sort=True)
    """

    fn_name = "join"

    other_expr: Optional[RawCode] = evals_into("other")
    on_expr: Optional[RawCode] = evals_into("on")
    how_expr: Optional[RawCode] = evals_into("how")
    sort_expr: Optional[RawCode] = evals_into("sort")


##############################################################################
# boolean expressions
##############################################################################


@dataclasses.dataclass
class BoolExprStep(ChainStep):
    """
    matches top-level boolean expressions.
    as a heuristic, we assume that the left side of the expr is the only thing
    we care about visualizing. so these cases will visualize properly:

    df['Sex'] == 'M'
    (df['Count'] > 13000) & (df['Count'] < 15000)

    but not these:

    'M' == df['Sex']
    (13000 < df['Count']) & (15000 > df['Count'])
    """

    pass


##############################################################################
# Subscripts
##############################################################################


@dataclasses.dataclass
class Subscript(ChainStep):
    """
    hard-codes two slice elements (i've never seen a third slice in pandas
    code, but i could be wrong)
    """

    slicer: Slicer

    slice1: SubscriptEl
    slice2: Optional[SubscriptEl]


@dataclasses.dataclass
class SubsSlice(Base):
    """df.iloc[:3]"""

    pass


@dataclasses.dataclass
class SubsComparison(Base):
    """
    special case for boolean expressions:

        df[df['Count'] > 10]

        col = 'Name'
        df[(df[col] > 10) | (df['Year'] >= 2020)]

    but won't handle cases where the expression runs outside the slide, like:

        mask = df['Count'] > 10
        df[mask]

    (those will go into EvalSlice)
    """

    # this is a list of code expressions that will each eval to one-or-more
    # labels. when parsing, we need to pull these out of each boolean mask
    label_exprs: List[RawCode] = field(
        metadata=dict(evals_into="{attr}_filter_labels")
    )


@dataclasses.dataclass
class SubsEval(Base):
    """
    anything that evals into row/column label(s), like:

        df['Name']
        df[['Name', 'Count']]
        df[df.columns[:4]]

    the evaluated expression can technically be any valid pandas slice,
    like boolean masks that aren't caught by ComparisonSlice.
    if the result isn't a list of labels, then we should just pass it through
    and not try to visualize it
    """

    expr: RawCode = field(metadata=dict(evals_into="{attr}_values"))


SubscriptEl = Union[SubsSlice, SubsComparison, SubsEval]

##############################################################################
# Errors
##############################################################################


@dataclasses.dataclass
class ParseSyntaxError(ChainStep):
    """
    represents an error in parsing. when this happens, we should pass along the
    error for serializing. we don't run the code since libcst produces nicer
    error messages compared to Python
    """

    error_msg: str


@dataclasses.dataclass
class EvalError(ChainStep):
    """represents a step in the chain that caused a runtime error"""

    # TODO: compute code positions for errors
    @classmethod
    def from_node(cls, node: Base):
        return cls(code=node.code, location=node.location)


ParseResult = Union[ParsedModule, ParseSyntaxError]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import dataclasses
import re
from pathlib import Path
from typing import Collection, List, Mapping, Optional, Pattern, Union

import libcst as cst
from libcst._metadata_dependent import MetadataDependent
from libcst.metadata.base_provider import BatchableMetadataProvider
from libcst.metadata.scope_provider import (
    QualifiedName,
    QualifiedNameSource,
    ScopeProvider,
)


class QualifiedNameProvider(BatchableMetadataProvider[Collection[QualifiedName]]):
    """
    Compute possible qualified names of a variable CSTNode
    (extends `PEP-3155 <https://www.python.org/dev/peps/pep-3155/>`_).
    It uses the
    :func:`~libcst.metadata.Scope.get_qualified_names_for` underlying to get qualified names.
    Multiple qualified names may be returned, such as when we have conditional imports or an
    import shadows another. E.g., the provider finds ``a.b``, ``d.e`` and
    ``f.g`` as possible qualified names of ``c``::

        >>> wrapper = MetadataWrapper(
        >>>     cst.parse_module(dedent(
        >>>     '''
        >>>         if something:
        >>>             from a import b as c
        >>>         elif otherthing:
        >>>             from d import e as c
        >>>         else:
        >>>             from f import g as c
        >>>         c()
        >>>     '''
        >>>     ))
        >>> )
        >>> call = wrapper.module.body[1].body[0].value
        >>> wrapper.resolve(QualifiedNameProvider)[call],
        {
            QualifiedName(name="a.b", source=QualifiedNameSource.IMPORT),
            QualifiedName(name="d.e", source=QualifiedNameSource.IMPORT),
            QualifiedName(name="f.g", source=QualifiedNameSource.IMPORT),
        }

    For qualified name of a variable in a function or a comprehension, please refer
    :func:`~libcst.metadata.Scope.get_qualified_names_for` for more detail.
    """

    METADATA_DEPENDENCIES = (ScopeProvider,)

    def visit_Module(self, node: cst.Module) -> Optional[bool]:
        visitor = QualifiedNameVisitor(self)
        node.visit(visitor)

    @staticmethod
    def has_name(
        visitor: MetadataDependent, node: cst.CSTNode, name: Union[str, QualifiedName]
    ) -> bool:
        """Check if any of qualified name has the str name or :class:`~libcst.metadata.QualifiedName` name."""
        qualified_names = visitor.get_metadata(QualifiedNameProvider, node, set())
        if isinstance(name, str):
            return any(qn.name == name for qn in qualified_names)
        else:
            return any(qn == name for qn in qualified_names)


class QualifiedNameVisitor(cst.CSTVisitor):
    def __init__(self, provider: "QualifiedNameProvider") -> None:
        self.provider: QualifiedNameProvider = provider

    def on_visit(self, node: cst.CSTNode) -> bool:
        scope = self.provider.get_metadata(ScopeProvider, node, None)
        if scope:
            self.provider.set_metadata(node, scope.get_qualified_names_for(node))
        else:
            self.provider.set_metadata(node, set())
        super().on_visit(node)
        return True


DOT_PY: Pattern[str] = re.compile(r"(__init__)?\.py$")


def _module_name(path: str) -> Optional[str]:
    return DOT_PY.sub("", path).replace("/", ".").rstrip(".")


class FullyQualifiedNameProvider(BatchableMetadataProvider[Collection[QualifiedName]]):
    """
    Provide fully qualified names for CST nodes. Like :class:`QualifiedNameProvider`,
    but the provided :class:`QualifiedName` instances have absolute identifier names
    instead of local to the current module.

    This provider is initialized with the current module's fully qualified name, and can
    be used with :class:`~libcst.metadata.FullRepoManager`. The module's fully qualified
    name itself is stored as a metadata of the :class:`~libcst.Module` node. Compared to
    :class:`QualifiedNameProvider`, it also resolves relative imports.

    Example usage::

        >>> mgr = FullRepoManager(".", {"dir/a.py"}, {FullyQualifiedNameProvider})
        >>> wrapper = mgr.get_metadata_wrapper_for_path("dir/a.py")
        >>> fqnames = wrapper.resolve(FullyQualifiedNameProvider)
        >>> {type(k): v for (k, v) in fqnames.items()}
        {<class 'libcst._nodes.module.Module'>: {QualifiedName(name='dir.a', source=<QualifiedNameSource.LOCAL: 3>)}}

    """

    METADATA_DEPENDENCIES = (QualifiedNameProvider,)

    @classmethod
    def gen_cache(
        cls, root_path: Path, paths: List[str], timeout: Optional[int] = None
    ) -> Mapping[str, object]:
        cache = {path: _module_name(path) for path in paths}
        return cache

    def __init__(self, cache: str) -> None:
        super().__init__(cache)
        self.module_name: str = cache

    def visit_Module(self, node: cst.Module) -> bool:
        visitor = FullyQualifiedNameVisitor(self, self.module_name)
        node.visit(visitor)
        self.set_metadata(
            node,
            {QualifiedName(name=self.module_name, source=QualifiedNameSource.LOCAL)},
        )
        return True


class FullyQualifiedNameVisitor(cst.CSTVisitor):
    @staticmethod
    def _fully_qualify_local(module_name: str, qname: QualifiedName) -> str:
        name = qname.name
        if not name.startswith("."):
            # not a relative import
            return f"{module_name}.{name}"

        # relative import
        name = name.lstrip(".")
        parts_to_strip = len(qname.name) - len(name)
        target_module = ".".join(module_name.split(".")[: -1 * parts_to_strip])
        return f"{target_module}.{name}"

    @staticmethod
    def _fully_qualify(module_name: str, qname: QualifiedName) -> QualifiedName:
        if qname.source == QualifiedNameSource.BUILTIN:
            # builtins are already fully qualified
            return qname
        name = qname.name
        if qname.source == QualifiedNameSource.IMPORT and not name.startswith("."):
            # non-relative imports are already fully qualified
            return qname
        new_name = FullyQualifiedNameVisitor._fully_qualify_local(module_name, qname)
        return dataclasses.replace(qname, name=new_name)

    def __init__(self, provider: FullyQualifiedNameProvider, module_name: str) -> None:
        self.module_name = module_name
        self.provider = provider

    def on_visit(self, node: cst.CSTNode) -> bool:
        qnames = self.provider.get_metadata(QualifiedNameProvider, node)
        if qnames is not None:
            self.provider.set_metadata(
                node,
                {
                    FullyQualifiedNameVisitor._fully_qualify(self.module_name, qname)
                    for qname in qnames
                },
            )
        return True



import pandas as pd
import io

csv = '''
Name,Sex,Count,Year
Noah,M,18252,2020
Julius,M,960,2020
Karen,M,6,2020
Karen,F,325,2020
Noah,F,305,2020
'''

csv2 = '''
nyt_name,category
Karen,boomer
Julius,mythology
Freya,mythology
'''

baby = pd.read_csv(io.StringIO(csv))
nyt = pd.read_csv(io.StringIO(csv2))

nyt2 = nyt.rename(columns={'category': 'Count'})

# overlapping columns get _x and _y appended
baby.merge(nyt2, left_on='Name', right_on='nyt_name')


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


# This file was generated by libcst.codegen.gen_type_mapping
from typing import Dict as TypingDict, Type, Union

from libcst._maybe_sentinel import MaybeSentinel
from libcst._nodes.base import CSTNode
from libcst._nodes.expression import (
    Annotation,
    Arg,
    Asynchronous,
    Attribute,
    Await,
    BaseDictElement,
    BaseElement,
    BaseExpression,
    BaseFormattedStringContent,
    BaseSlice,
    BinaryOperation,
    BooleanOperation,
    Call,
    CompFor,
    CompIf,
    Comparison,
    ComparisonTarget,
    ConcatenatedString,
    Dict,
    DictComp,
    DictElement,
    Element,
    Ellipsis,
    Float,
    FormattedString,
    FormattedStringExpression,
    FormattedStringText,
    From,
    GeneratorExp,
    IfExp,
    Imaginary,
    Index,
    Integer,
    Lambda,
    LeftCurlyBrace,
    LeftParen,
    LeftSquareBracket,
    List,
    ListComp,
    Name,
    NamedExpr,
    Param,
    ParamSlash,
    ParamStar,
    Parameters,
    RightCurlyBrace,
    RightParen,
    RightSquareBracket,
    Set,
    SetComp,
    SimpleString,
    Slice,
    StarredDictElement,
    StarredElement,
    Subscript,
    SubscriptElement,
    Tuple,
    UnaryOperation,
    Yield,
)
from libcst._nodes.module import Module
from libcst._nodes.op import (
    Add,
    AddAssign,
    And,
    AssignEqual,
    BaseAugOp,
    BaseBinaryOp,
    BaseBooleanOp,
    BaseCompOp,
    BaseUnaryOp,
    BitAnd,
    BitAndAssign,
    BitInvert,
    BitOr,
    BitOrAssign,
    BitXor,
    BitXorAssign,
    Colon,
    Comma,
    Divide,
    DivideAssign,
    Dot,
    Equal,
    FloorDivide,
    FloorDivideAssign,
    GreaterThan,
    GreaterThanEqual,
    ImportStar,
    In,
    Is,
    IsNot,
    LeftShift,
    LeftShiftAssign,
    LessThan,
    LessThanEqual,
    MatrixMultiply,
    MatrixMultiplyAssign,
    Minus,
    Modulo,
    ModuloAssign,
    Multiply,
    MultiplyAssign,
    Not,
    NotEqual,
    NotIn,
    Or,
    Plus,
    Power,
    PowerAssign,
    RightShift,
    RightShiftAssign,
    Semicolon,
    Subtract,
    SubtractAssign,
)
from libcst._nodes.statement import (
    AnnAssign,
    AsName,
    Assert,
    Assign,
    AssignTarget,
    AugAssign,
    BaseSmallStatement,
    BaseStatement,
    BaseSuite,
    Break,
    ClassDef,
    Continue,
    Decorator,
    Del,
    Else,
    ExceptHandler,
    Expr,
    Finally,
    For,
    FunctionDef,
    Global,
    If,
    Import,
    ImportAlias,
    ImportFrom,
    IndentedBlock,
    NameItem,
    Nonlocal,
    Pass,
    Raise,
    Return,
    SimpleStatementLine,
    SimpleStatementSuite,
    Try,
    While,
    With,
    WithItem,
)
from libcst._nodes.whitespace import (
    BaseParenthesizableWhitespace,
    Comment,
    EmptyLine,
    Newline,
    ParenthesizedWhitespace,
    SimpleWhitespace,
    TrailingWhitespace,
)
from libcst._removal_sentinel import RemovalSentinel


TYPED_FUNCTION_RETURN_MAPPING: TypingDict[Type[CSTNode], object] = {
    Add: BaseBinaryOp,
    AddAssign: BaseAugOp,
    And: BaseBooleanOp,
    AnnAssign: Union[BaseSmallStatement, RemovalSentinel],
    Annotation: Annotation,
    Arg: Union[Arg, RemovalSentinel],
    AsName: AsName,
    Assert: Union[BaseSmallStatement, RemovalSentinel],
    Assign: Union[BaseSmallStatement, RemovalSentinel],
    AssignEqual: Union[AssignEqual, MaybeSentinel],
    AssignTarget: Union[AssignTarget, RemovalSentinel],
    Asynchronous: Asynchronous,
    Attribute: BaseExpression,
    AugAssign: Union[BaseSmallStatement, RemovalSentinel],
    Await: BaseExpression,
    BinaryOperation: BaseExpression,
    BitAnd: BaseBinaryOp,
    BitAndAssign: BaseAugOp,
    BitInvert: BaseUnaryOp,
    BitOr: BaseBinaryOp,
    BitOrAssign: BaseAugOp,
    BitXor: BaseBinaryOp,
    BitXorAssign: BaseAugOp,
    BooleanOperation: BaseExpression,
    Break: Union[BaseSmallStatement, RemovalSentinel],
    Call: BaseExpression,
    ClassDef: Union[BaseStatement, RemovalSentinel],
    Colon: Union[Colon, MaybeSentinel],
    Comma: Union[Comma, MaybeSentinel],
    Comment: Comment,
    CompFor: CompFor,
    CompIf: CompIf,
    Comparison: BaseExpression,
    ComparisonTarget: Union[ComparisonTarget, RemovalSentinel],
    ConcatenatedString: BaseExpression,
    Continue: Union[BaseSmallStatement, RemovalSentinel],
    Decorator: Union[Decorator, RemovalSentinel],
    Del: Union[BaseSmallStatement, RemovalSentinel],
    Dict: BaseExpression,
    DictComp: BaseExpression,
    DictElement: Union[BaseDictElement, RemovalSentinel],
    Divide: BaseBinaryOp,
    DivideAssign: BaseAugOp,
    Dot: Union[Dot, RemovalSentinel],
    Element: Union[BaseElement, RemovalSentinel],
    Ellipsis: BaseExpression,
    Else: Else,
    EmptyLine: Union[EmptyLine, RemovalSentinel],
    Equal: BaseCompOp,
    ExceptHandler: Union[ExceptHandler, RemovalSentinel],
    Expr: Union[BaseSmallStatement, RemovalSentinel],
    Finally: Finally,
    Float: BaseExpression,
    FloorDivide: BaseBinaryOp,
    FloorDivideAssign: BaseAugOp,
    For: Union[BaseStatement, RemovalSentinel],
    FormattedString: BaseExpression,
    FormattedStringExpression: Union[BaseFormattedStringContent, RemovalSentinel],
    FormattedStringText: Union[BaseFormattedStringContent, RemovalSentinel],
    From: From,
    FunctionDef: Union[BaseStatement, RemovalSentinel],
    GeneratorExp: BaseExpression,
    Global: Union[BaseSmallStatement, RemovalSentinel],
    GreaterThan: BaseCompOp,
    GreaterThanEqual: BaseCompOp,
    If: Union[BaseStatement, RemovalSentinel],
    IfExp: BaseExpression,
    Imaginary: BaseExpression,
    Import: Union[BaseSmallStatement, RemovalSentinel],
    ImportAlias: Union[ImportAlias, RemovalSentinel],
    ImportFrom: Union[BaseSmallStatement, RemovalSentinel],
    ImportStar: ImportStar,
    In: BaseCompOp,
    IndentedBlock: BaseSuite,
    Index: BaseSlice,
    Integer: BaseExpression,
    Is: BaseCompOp,
    IsNot: BaseCompOp,
    Lambda: BaseExpression,
    LeftCurlyBrace: LeftCurlyBrace,
    LeftParen: Union[LeftParen, MaybeSentinel, RemovalSentinel],
    LeftShift: BaseBinaryOp,
    LeftShiftAssign: BaseAugOp,
    LeftSquareBracket: LeftSquareBracket,
    LessThan: BaseCompOp,
    LessThanEqual: BaseCompOp,
    List: BaseExpression,
    ListComp: BaseExpression,
    MatrixMultiply: BaseBinaryOp,
    MatrixMultiplyAssign: BaseAugOp,
    Minus: BaseUnaryOp,
    Module: Module,
    Modulo: BaseBinaryOp,
    ModuloAssign: BaseAugOp,
    Multiply: BaseBinaryOp,
    MultiplyAssign: BaseAugOp,
    Name: BaseExpression,
    NameItem: Union[NameItem, RemovalSentinel],
    NamedExpr: BaseExpression,
    Newline: Newline,
    Nonlocal: Union[BaseSmallStatement, RemovalSentinel],
    Not: BaseUnaryOp,
    NotEqual: BaseCompOp,
    NotIn: BaseCompOp,
    Or: BaseBooleanOp,
    Param: Union[Param, MaybeSentinel, RemovalSentinel],
    ParamSlash: Union[ParamSlash, MaybeSentinel],
    ParamStar: Union[ParamStar, MaybeSentinel],
    Parameters: Parameters,
    ParenthesizedWhitespace: Union[BaseParenthesizableWhitespace, MaybeSentinel],
    Pass: Union[BaseSmallStatement, RemovalSentinel],
    Plus: BaseUnaryOp,
    Power: BaseBinaryOp,
    PowerAssign: BaseAugOp,
    Raise: Union[BaseSmallStatement, RemovalSentinel],
    Return: Union[BaseSmallStatement, RemovalSentinel],
    RightCurlyBrace: RightCurlyBrace,
    RightParen: Union[RightParen, MaybeSentinel, RemovalSentinel],
    RightShift: BaseBinaryOp,
    RightShiftAssign: BaseAugOp,
    RightSquareBracket: RightSquareBracket,
    Semicolon: Union[Semicolon, MaybeSentinel],
    Set: BaseExpression,
    SetComp: BaseExpression,
    SimpleStatementLine: Union[BaseStatement, RemovalSentinel],
    SimpleStatementSuite: BaseSuite,
    SimpleString: BaseExpression,
    SimpleWhitespace: Union[BaseParenthesizableWhitespace, MaybeSentinel],
    Slice: BaseSlice,
    StarredDictElement: Union[BaseDictElement, RemovalSentinel],
    StarredElement: Union[BaseElement, RemovalSentinel],
    Subscript: BaseExpression,
    SubscriptElement: Union[SubscriptElement, RemovalSentinel],
    Subtract: BaseBinaryOp,
    SubtractAssign: BaseAugOp,
    TrailingWhitespace: TrailingWhitespace,
    Try: Union[BaseStatement, RemovalSentinel],
    Tuple: BaseExpression,
    UnaryOperation: BaseExpression,
    While: Union[BaseStatement, RemovalSentinel],
    With: Union[BaseStatement, RemovalSentinel],
    WithItem: Union[WithItem, RemovalSentinel],
    Yield: BaseExpression,
}



"""
Provide saving, loading and presenting gathered `ClassTracker` statistics.
"""

import os
import sys
from pympler.util.compat import pickle
from copy import deepcopy
from pympler.util.stringutils import trunc, pp, pp_timestamp

from pympler.asizeof import Asized


__all__ = ["Stats", "ConsoleStats", "HtmlStats"]


def _merge_asized(base, other, level=0):
    """
    Merge **Asized** instances `base` and `other` into `base`.
    """
    ref2key = lambda ref: ref.name.split(':')[0]
    base.size += other.size
    base.flat += other.flat
    if level > 0:
        base.name = ref2key(base)
    # Add refs from other to base. Any new refs are appended.
    base.refs = list(base.refs)  # we may need to append items
    refs = {}
    for ref in base.refs:
        refs[ref2key(ref)] = ref
    for ref in other.refs:
        key = ref2key(ref)
        if key in refs:
            _merge_asized(refs[key], ref, level=level + 1)
        else:
            # Don't modify existing Asized instances => deepcopy
            base.refs.append(deepcopy(ref))
            base.refs[-1].name = key


def _merge_objects(tref, merged, obj):
    """
    Merge the snapshot size information of multiple tracked objects.  The
    tracked object `obj` is scanned for size information at time `tref`.
    The sizes are merged into **Asized** instance `merged`.
    """
    size = None
    for (timestamp, tsize) in obj.snapshots:
        if timestamp == tref:
            size = tsize
    if size:
        _merge_asized(merged, size)


def _format_trace(trace):
    """
    Convert the (stripped) stack-trace to a nice readable format. The stack
    trace `trace` is a list of frame records as returned by
    **inspect.stack** but without the frame objects.
    Returns a string.
    """
    lines = []
    for fname, lineno, func, src, _ in trace:
        if src:
            for line in src:
                lines.append('    ' + line.strip() + '\n')
        lines.append('  %s:%4d in %s\n' % (fname, lineno, func))
    return ''.join(lines)


class Stats(object):
    """
    Presents the memory statistics gathered by a `ClassTracker` based on user
    preferences.
    """

    def __init__(self, tracker=None, filename=None, stream=None):
        """
        Initialize the data log structures either from a `ClassTracker`
        instance (argument `tracker`) or a previously dumped file (argument
        `filename`).

        :param tracker: ClassTracker instance
        :param filename: filename of previously dumped statistics
        :param stream: where to print statistics, defaults to ``sys.stdout``
        """
        if stream:
            self.stream = stream
        else:
            self.stream = sys.stdout
        self.tracker = tracker
        if tracker:
            self.index = tracker.index
            self.snapshots = tracker.snapshots
            self.history = tracker.history
        else:
            self.index = None
            self.snapshots = None
        self.sorted = []
        if filename:
            self.load_stats(filename)

    def load_stats(self, fdump):
        """
        Load the data from a dump file.
        The argument `fdump` can be either a filename or an open file object
        that requires read access.
        """
        if isinstance(fdump, type('')):
            fdump = open(fdump, 'rb')
        self.index = pickle.load(fdump)
        self.snapshots = pickle.load(fdump)
        self.sorted = []

    def dump_stats(self, fdump, close=True):
        """
        Dump the logged data to a file.
        The argument `file` can be either a filename or an open file object
        that requires write access. `close` controls if the file is closed
        before leaving this method (the default behaviour).
        """
        if self.tracker:
            self.tracker.stop_periodic_snapshots()

        if isinstance(fdump, type('')):
            fdump = open(fdump, 'wb')
        pickle.dump(self.index, fdump, protocol=pickle.HIGHEST_PROTOCOL)
        pickle.dump(self.snapshots, fdump, protocol=pickle.HIGHEST_PROTOCOL)
        if close:
            fdump.close()

    def _init_sort(self):
        """
        Prepare the data to be sorted.
        If not yet sorted, import all tracked objects from the tracked index.
        Extend the tracking information by implicit information to make
        sorting easier (DSU pattern).
        """
        if not self.sorted:
            # Identify the snapshot that tracked the largest amount of memory.
            tmax = None
            maxsize = 0
            for snapshot in self.snapshots:
                if snapshot.tracked_total > maxsize:
                    tmax = snapshot.timestamp
            for key in list(self.index.keys()):
                for tobj in self.index[key]:
                    tobj.classname = key
                    tobj.size = tobj.get_max_size()
                    tobj.tsize = tobj.get_size_at_time(tmax)
                self.sorted.extend(self.index[key])

    def sort_stats(self, *args):
        """
        Sort the tracked objects according to the supplied criteria. The
        argument is a string identifying the basis of a sort (example: 'size'
        or 'classname'). When more than one key is provided, then additional
        keys are used as secondary criteria when there is equality in all keys
        selected before them. For example, ``sort_stats('name', 'size')`` will
        sort all the entries according to their class name, and resolve all
        ties (identical class names) by sorting by size.  The criteria are
        fields in the tracked object instances. Results are stored in the
        ``self.sorted`` list which is used by ``Stats.print_stats()`` and other
        methods. The fields available for sorting are:

            'classname'
                the name with which the class was registered
            'name'
                the classname
            'birth'
                creation timestamp
            'death'
                destruction timestamp
            'size'
                the maximum measured size of the object
            'tsize'
                the measured size during the largest snapshot
            'repr'
                string representation of the object

        Note that sorts on size are in descending order (placing most memory
        consuming items first), whereas name, repr, and creation time searches
        are in ascending order (alphabetical).

        The function returns self to allow calling functions on the result::

            stats.sort_stats('size').reverse_order().print_stats()
        """

        criteria = ('classname', 'tsize', 'birth', 'death',
                    'name', 'repr', 'size')

        if not set(criteria).issuperset(set(args)):
            raise ValueError("Invalid sort criteria")

        if not args:
            args = criteria

        def args_to_tuple(obj):
            keys = []
            for attr in args:
                attribute = getattr(obj, attr)
                if attr in ('tsize', 'size'):
                    attribute = -attribute
                keys.append(attribute)
            return tuple(keys)

        self._init_sort()
        self.sorted.sort(key=args_to_tuple)

        return self

    def reverse_order(self):
        """
        Reverse the order of the tracked instance index `self.sorted`.
        """
        self._init_sort()
        self.sorted.reverse()
        return self

    def annotate(self):
        """
        Annotate all snapshots with class-based summaries.
        """
        for snapshot in self.snapshots:
            self.annotate_snapshot(snapshot)

    def annotate_snapshot(self, snapshot):
        """
        Store additional statistical data in snapshot.
        """
        if hasattr(snapshot, 'classes'):
            return

        snapshot.classes = {}

        for classname in list(self.index.keys()):
            total = 0
            active = 0
            merged = Asized(0, 0)
            for tobj in self.index[classname]:
                _merge_objects(snapshot.timestamp, merged, tobj)
                total += tobj.get_size_at_time(snapshot.timestamp)
                if (tobj.birth < snapshot.timestamp and
                        (tobj.death is None or
                         tobj.death > snapshot.timestamp)):
                    active += 1

            try:
                pct = total * 100.0 / snapshot.total
            except ZeroDivisionError:  # pragma: no cover
                pct = 0
            try:
                avg = total / active
            except ZeroDivisionError:
                avg = 0

            snapshot.classes[classname] = dict(sum=total,
                                               avg=avg,
                                               pct=pct,
                                               active=active)
            snapshot.classes[classname]['merged'] = merged

    @property
    def tracked_classes(self):
        """Return a list of all tracked classes occurring in any snapshot."""
        return sorted(list(self.index.keys()))


class ConsoleStats(Stats):
    """
    Presentation layer for `Stats` to be used in text-based consoles.
    """

    def _print_refs(self, refs, total, prefix='    ',
                    level=1, minsize=0, minpct=0.1):
        """
        Print individual referents recursively.
        """
        lrefs = list(refs)
        lrefs.sort(key=lambda x: x.size)
        lrefs.reverse()
        for ref in lrefs:
            if ref.size > minsize and (ref.size * 100.0 / total) > minpct:
                self.stream.write('%-50s %-14s %3d%% [%d]\n' % (
                    trunc(prefix + str(ref.name), 50),
                    pp(ref.size),
                    int(ref.size * 100.0 / total),
                    level
                ))
                self._print_refs(ref.refs, total, prefix=prefix + '  ',
                                 level=level + 1)

    def print_object(self, tobj):
        """
        Print the gathered information of object `tobj` in human-readable
        format.
        """
        if tobj.death:
            self.stream.write('%-32s ( free )   %-35s\n' % (
                trunc(tobj.name, 32, left=1), trunc(tobj.repr, 35)))
        else:
            self.stream.write('%-32s 0x%08x %-35s\n' % (
                trunc(tobj.name, 32, left=1),
                tobj.id,
                trunc(tobj.repr, 35)
            ))
        if tobj.trace:
            self.stream.write(_format_trace(tobj.trace))
        for (timestamp, size) in tobj.snapshots:
            self.stream.write('  %-30s %s\n' % (
                pp_timestamp(timestamp), pp(size.size)
            ))
            self._print_refs(size.refs, size.size)
        if tobj.death is not None:
            self.stream.write('  %-30s finalize\n' % (
                pp_timestamp(tobj.death),
            ))

    def print_stats(self, clsname=None, limit=1.0):
        """
        Write tracked objects to stdout.  The output can be filtered and
        pruned.  Only objects are printed whose classname contain the substring
        supplied by the `clsname` argument.  The output can be pruned by
        passing a `limit` value.

        :param clsname: Only print objects whose classname contain the given
            substring.
        :param limit: If `limit` is a float smaller than one, only the supplied
            percentage of the total tracked data is printed. If `limit` is
            bigger than one, this number of tracked objects are printed.
            Tracked objects are first filtered, and then pruned (if specified).
        """
        if self.tracker:
            self.tracker.stop_periodic_snapshots()

        if not self.sorted:
            self.sort_stats()

        _sorted = self.sorted

        if clsname:
            _sorted = [to for to in _sorted if clsname in to.classname]

        if limit < 1.0:
            limit = max(1, int(len(self.sorted) * limit))
        _sorted = _sorted[:int(limit)]

        # Emit per-instance data
        for tobj in _sorted:
            self.print_object(tobj)

    def print_summary(self):
        """
        Print per-class summary for each snapshot.
        """
        # Emit class summaries for each snapshot
        classlist = self.tracked_classes

        fobj = self.stream

        fobj.write('---- SUMMARY ' + '-' * 66 + '\n')
        for snapshot in self.snapshots:
            self.annotate_snapshot(snapshot)
            fobj.write('%-35s %11s %12s %12s %5s\n' % (
                trunc(snapshot.desc, 35),
                'active',
                pp(snapshot.asizeof_total),
                'average',
                'pct'
            ))
            for classname in classlist:
                info = snapshot.classes.get(classname)
                fobj.write('  %-33s %11d %12s %12s %4d%%\n' % (
                    trunc(classname, 33),
                    info['active'],
                    pp(info['sum']),
                    pp(info['avg']),
                    info['pct']
                ))
        fobj.write('-' * 79 + '\n')


class HtmlStats(Stats):
    """
    Output the `ClassTracker` statistics as HTML pages and graphs.
    """

    style = """<style type="text/css">
        table { width:100%; border:1px solid #000; border-spacing:0px; }
        td, th { border:0px; }
        div { width:200px; padding:10px; background-color:#FFEECC; }
        #nb { border:0px; }
        #tl { margin-top:5mm; margin-bottom:5mm; }
        #p1 { padding-left: 5px; }
        #p2 { padding-left: 50px; }
        #p3 { padding-left: 100px; }
        #p4 { padding-left: 150px; }
        #p5 { padding-left: 200px; }
        #p6 { padding-left: 210px; }
        #p7 { padding-left: 220px; }
        #hl { background-color:#FFFFCC; }
        #r1 { background-color:#BBBBBB; }
        #r2 { background-color:#CCCCCC; }
        #r3 { background-color:#DDDDDD; }
        #r4 { background-color:#EEEEEE; }
        #r5,#r6,#r7 { background-color:#FFFFFF; }
        #num { text-align:right; }
    </style>
    """

    nopylab_msg = """<div color="#FFCCCC">Could not generate %s chart!
    Install <a href="http://matplotlib.sourceforge.net/">Matplotlib</a>
    to generate charts.</div>\n"""

    chart_tag = '<img src="%s">\n'
    header = "<html><head><title>%s</title>%s</head><body>\n"
    tableheader = '<table border="1">\n'
    tablefooter = '</table>\n'
    footer = '</body></html>\n'

    refrow = """<tr id="r%(level)d">
        <td id="p%(level)d">%(name)s</td>
        <td id="num">%(size)s</td>
        <td id="num">%(pct)3.1f%%</td></tr>"""

    def _print_refs(self, fobj, refs, total, level=1, minsize=0, minpct=0.1):
        """
        Print individual referents recursively.
        """
        lrefs = list(refs)
        lrefs.sort(key=lambda x: x.size)
        lrefs.reverse()
        if level == 1:
            fobj.write('<table>\n')
        for ref in lrefs:
            if ref.size > minsize and (ref.size * 100.0 / total) > minpct:
                data = dict(level=level,
                            name=trunc(str(ref.name), 128),
                            size=pp(ref.size),
                            pct=ref.size * 100.0 / total)
                fobj.write(self.refrow % data)
                self._print_refs(fobj, ref.refs, total, level=level + 1)
        if level == 1:
            fobj.write("</table>\n")

    class_summary = """<p>%(cnt)d instances of %(cls)s were registered. The
        average size is %(avg)s, the minimal size is %(min)s, the maximum size
        is %(max)s.</p>\n"""
    class_snapshot = '''<h3>Snapshot: %(name)s, %(total)s occupied by instances
        of class %(cls)s</h3>\n'''

    def print_class_details(self, fname, classname):
        """
        Print detailed statistics and instances for the class `classname`. All
        data will be written to the file `fname`.
        """
        fobj = open(fname, "w")
        fobj.write(self.header % (classname, self.style))

        fobj.write("<h1>%s</h1>\n" % (classname))

        sizes = [tobj.get_max_size() for tobj in self.index[classname]]
        total = 0
        for s in sizes:
            total += s
        data = {'cnt': len(self.index[classname]), 'cls': classname}
        data['avg'] = pp(total / len(sizes))
        data['max'] = pp(max(sizes))
        data['min'] = pp(min(sizes))
        fobj.write(self.class_summary % data)

        fobj.write(self.charts[classname])

        fobj.write("<h2>Coalesced Referents per Snapshot</h2>\n")
        for snapshot in self.snapshots:
            if classname in snapshot.classes:
                merged = snapshot.classes[classname]['merged']
                fobj.write(self.class_snapshot % {
                    'name': snapshot.desc,
                    'cls': classname,
                    'total': pp(merged.size),
                })
                if merged.refs:
                    self._print_refs(fobj, merged.refs, merged.size)
                else:
                    fobj.write('<p>No per-referent sizes recorded.</p>\n')

        fobj.write("<h2>Instances</h2>\n")
        for tobj in self.index[classname]:
            fobj.write('<table id="tl" width="100%" rules="rows">\n')
            fobj.write('<tr><td id="hl" width="140px">Instance</td>' +
                       '<td id="hl">%s at 0x%08x</td></tr>\n' %
                       (tobj.name, tobj.id))
            if tobj.repr:
                fobj.write("<tr><td>Representation</td>" +
                           "<td>%s&nbsp;</td></tr>\n" % tobj.repr)
            fobj.write("<tr><td>Lifetime</td><td>%s - %s</td></tr>\n" %
                       (pp_timestamp(tobj.birth), pp_timestamp(tobj.death)))
            if tobj.trace:
                trace = "<pre>%s</pre>" % (_format_trace(tobj.trace))
                fobj.write("<tr><td>Instantiation</td><td>%s</td></tr>\n" %
                           trace)
            for (timestamp, size) in tobj.snapshots:
                fobj.write("<tr><td>%s</td>" % pp_timestamp(timestamp))
                if not size.refs:
                    fobj.write("<td>%s</td></tr>\n" % pp(size.size))
                else:
                    fobj.write("<td>%s" % pp(size.size))
                    self._print_refs(fobj, size.refs, size.size)
                    fobj.write("</td></tr>\n")
            fobj.write("</table>\n")

        fobj.write(self.footer)
        fobj.close()

    snapshot_cls_header = """<tr>
        <th id="hl">Class</th>
        <th id="hl" align="right">Instance #</th>
        <th id="hl" align="right">Total</th>
        <th id="hl" align="right">Average size</th>
        <th id="hl" align="right">Share</th></tr>\n"""

    snapshot_cls = """<tr>
        <td>%(cls)s</td>
        <td align="right">%(active)d</td>
        <td align="right">%(sum)s</td>
        <td align="right">%(avg)s</td>
        <td align="right">%(pct)3.2f%%</td></tr>\n"""

    snapshot_summary = """<p>Total virtual memory assigned to the program
        at that time was %(sys)s, which includes %(overhead)s profiling
        overhead. The ClassTracker tracked %(tracked)s in total. The measurable
        objects including code objects but excluding overhead have a total size
        of %(asizeof)s.</p>\n"""

    def relative_path(self, filepath, basepath=None):
        """
        Convert the filepath path to a relative path against basepath. By
        default basepath is self.basedir.
        """
        if basepath is None:
            basepath = self.basedir
        if not basepath:
            return filepath
        if filepath.startswith(basepath):
            filepath = filepath[len(basepath):]
        if filepath and filepath[0] == os.sep:
            filepath = filepath[1:]
        return filepath

    def create_title_page(self, filename, title=''):
        """
        Output the title page.
        """
        fobj = open(filename, "w")
        fobj.write(self.header % (title, self.style))

        fobj.write("<h1>%s</h1>\n" % title)
        fobj.write("<h2>Memory distribution over time</h2>\n")
        fobj.write(self.charts['snapshots'])

        fobj.write("<h2>Snapshots statistics</h2>\n")
        fobj.write('<table id="nb">\n')

        classlist = list(self.index.keys())
        classlist.sort()

        for snapshot in self.snapshots:
            fobj.write('<tr><td>\n')
            fobj.write('<table id="tl" rules="rows">\n')
            fobj.write("<h3>%s snapshot at %s</h3>\n" % (
                snapshot.desc or 'Untitled',
                pp_timestamp(snapshot.timestamp)
            ))

            data = {}
            data['sys'] = pp(snapshot.system_total.vsz)
            data['tracked'] = pp(snapshot.tracked_total)
            data['asizeof'] = pp(snapshot.asizeof_total)
            data['overhead'] = pp(getattr(snapshot, 'overhead', 0))

            fobj.write(self.snapshot_summary % data)

            if snapshot.tracked_total:
                fobj.write(self.snapshot_cls_header)
                for classname in classlist:
                    data = snapshot.classes[classname].copy()
                    path = self.relative_path(self.links[classname])
                    data['cls'] = '<a href="%s">%s</a>' % (path, classname)
                    data['sum'] = pp(data['sum'])
                    data['avg'] = pp(data['avg'])
                    fobj.write(self.snapshot_cls % data)
            fobj.write('</table>')
            fobj.write('</td><td>\n')
            if snapshot.tracked_total:
                fobj.write(self.charts[snapshot])
            fobj.write('</td></tr>\n')

        fobj.write("</table>\n")
        fobj.write(self.footer)
        fobj.close()

    def create_lifetime_chart(self, classname, filename=''):
        """
        Create chart that depicts the lifetime of the instance registered with
        `classname`. The output is written to `filename`.
        """
        try:
            from pylab import figure, title, xlabel, ylabel, plot, savefig
        except ImportError:
            return HtmlStats.nopylab_msg % (classname + " lifetime")

        cnt = []
        for tobj in self.index[classname]:
            cnt.append([tobj.birth, 1])
            if tobj.death:
                cnt.append([tobj.death, -1])
        cnt.sort()
        for i in range(1, len(cnt)):
            cnt[i][1] += cnt[i - 1][1]

        x = [t for [t, c] in cnt]
        y = [c for [t, c] in cnt]

        figure()
        xlabel("Execution time [s]")
        ylabel("Instance #")
        title("%s instances" % classname)
        plot(x, y, 'o')
        savefig(filename)

        return self.chart_tag % (os.path.basename(filename))

    def create_snapshot_chart(self, filename=''):
        """
        Create chart that depicts the memory allocation over time apportioned
        to the tracked classes.
        """
        try:
            from pylab import (figure, title, xlabel, ylabel, plot, fill,
                               legend, savefig)
            import matplotlib.mlab as mlab
        except ImportError:
            return self.nopylab_msg % ("memory allocation")

        classlist = self.tracked_classes

        times = [snapshot.timestamp for snapshot in self.snapshots]
        base = [0] * len(self.snapshots)
        poly_labels = []
        polys = []
        for cn in classlist:
            pct = [snapshot.classes[cn]['pct'] for snapshot in self.snapshots]
            if pct and max(pct) > 3.0:
                sz = [float(fp.classes[cn]['sum']) / (1024 * 1024)
                      for fp in self.snapshots]
                sz = [sx + sy for sx, sy in zip(base, sz)]
                xp, yp = mlab.poly_between(times, base, sz)
                polys.append(((xp, yp), {'label': cn}))
                poly_labels.append(cn)
                base = sz

        figure()
        title("Snapshot Memory")
        xlabel("Execution Time [s]")
        ylabel("Virtual Memory [MiB]")

        sizes = [float(fp.asizeof_total) / (1024 * 1024)
                 for fp in self.snapshots]
        plot(times, sizes, 'r--', label='Total')
        sizes = [float(fp.tracked_total) / (1024 * 1024)
                 for fp in self.snapshots]
        plot(times, sizes, 'b--', label='Tracked total')

        for (args, kwds) in polys:
            fill(*args, **kwds)
        legend(loc=2)
        savefig(filename)

        return self.chart_tag % (self.relative_path(filename))

    def create_pie_chart(self, snapshot, filename=''):
        """
        Create a pie chart that depicts the distribution of the allocated
        memory for a given `snapshot`. The chart is saved to `filename`.
        """
        try:
            from pylab import figure, title, pie, axes, savefig
            from pylab import sum as pylab_sum
        except ImportError:
            return self.nopylab_msg % ("pie_chart")

        # Don't bother illustrating a pie without pieces.
        if not snapshot.tracked_total:
            return ''

        classlist = []
        sizelist = []
        for k, v in list(snapshot.classes.items()):
            if v['pct'] > 3.0:
                classlist.append(k)
                sizelist.append(v['sum'])
        sizelist.insert(0, snapshot.asizeof_total - pylab_sum(sizelist))
        classlist.insert(0, 'Other')
        #sizelist = [x*0.01 for x in sizelist]

        title("Snapshot (%s) Memory Distribution" % (snapshot.desc))
        figure(figsize=(8, 8))
        axes([0.1, 0.1, 0.8, 0.8])
        pie(sizelist, labels=classlist)
        savefig(filename, dpi=50)

        return self.chart_tag % (self.relative_path(filename))

    def create_html(self, fname, title="ClassTracker Statistics"):
        """
        Create HTML page `fname` and additional files in a directory derived
        from `fname`.
        """
        # Create a folder to store the charts and additional HTML files.
        self.basedir = os.path.dirname(os.path.abspath(fname))
        self.filesdir = os.path.splitext(fname)[0] + '_files'
        if not os.path.isdir(self.filesdir):
            os.mkdir(self.filesdir)
        self.filesdir = os.path.abspath(self.filesdir)
        self.links = {}

        # Annotate all snapshots in advance
        self.annotate()

        # Create charts. The tags to show the images are returned and stored in
        # the self.charts dictionary. This allows to return alternative text if
        # the chart creation framework is not available.
        self.charts = {}
        fn = os.path.join(self.filesdir, 'timespace.png')
        self.charts['snapshots'] = self.create_snapshot_chart(fn)

        for fp, idx in zip(self.snapshots, list(range(len(self.snapshots)))):
            fn = os.path.join(self.filesdir, 'fp%d.png' % (idx))
            self.charts[fp] = self.create_pie_chart(fp, fn)

        for cn in list(self.index.keys()):
            fn = os.path.join(self.filesdir, cn.replace('.', '_') + '-lt.png')
            self.charts[cn] = self.create_lifetime_chart(cn, fn)

        # Create HTML pages first for each class and then the index page.
        for cn in list(self.index.keys()):
            fn = os.path.join(self.filesdir, cn.replace('.', '_') + '.html')
            self.links[cn] = fn
            self.print_class_details(fn, cn)

        self.create_title_page(fname, title=title)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from dataclasses import dataclass
from typing import Generic, Optional, Sequence, TypeVar, Union

from libcst._add_slots import add_slots
from libcst._nodes.expression import (
    Annotation,
    Arg,
    Attribute,
    BaseExpression,
    BaseFormattedStringContent,
    Index,
    LeftParen,
    LeftSquareBracket,
    Name,
    Parameters,
    RightParen,
    RightSquareBracket,
    Slice,
    SubscriptElement,
)
from libcst._nodes.op import AssignEqual, BaseAugOp, Colon, Dot
from libcst._nodes.statement import AsName, BaseSmallStatement, Decorator, ImportAlias
from libcst._nodes.whitespace import EmptyLine, SimpleWhitespace, TrailingWhitespace
from libcst._parser.types.whitespace_state import WhitespaceState

_T = TypeVar("_T")


@add_slots
@dataclass(frozen=True)
class WithLeadingWhitespace(Generic[_T]):
    value: _T
    whitespace_before: WhitespaceState


@add_slots
@dataclass(frozen=True)
class SimpleStatementPartial:
    body: Sequence[BaseSmallStatement]
    whitespace_before: WhitespaceState
    trailing_whitespace: TrailingWhitespace


@add_slots
@dataclass(frozen=True)
class SlicePartial:
    second_colon: Colon
    step: Optional[BaseExpression]


@add_slots
@dataclass(frozen=True)
class AttributePartial:
    dot: Dot
    attr: Name


@add_slots
@dataclass(frozen=True)
class ArglistPartial:
    args: Sequence[Arg]


@add_slots
@dataclass(frozen=True)
class CallPartial:
    lpar: WithLeadingWhitespace[LeftParen]
    args: Sequence[Arg]
    rpar: RightParen


@add_slots
@dataclass(frozen=True)
class SubscriptPartial:
    slice: Union[Index, Slice, Sequence[SubscriptElement]]
    lbracket: LeftSquareBracket
    rbracket: RightSquareBracket
    whitespace_before: WhitespaceState


@add_slots
@dataclass(frozen=True)
class AnnAssignPartial:
    annotation: Annotation
    equal: Optional[AssignEqual]
    value: Optional[BaseExpression]


@add_slots
@dataclass(frozen=True)
class AugAssignPartial:
    operator: BaseAugOp
    value: BaseExpression


@add_slots
@dataclass(frozen=True)
class AssignPartial:
    equal: AssignEqual
    value: BaseExpression


class ParamStarPartial:
    pass


@add_slots
@dataclass(frozen=True)
class FuncdefPartial:
    lpar: LeftParen
    params: Parameters
    rpar: RightParen


@add_slots
@dataclass(frozen=True)
class DecoratorPartial:
    decorators: Sequence[Decorator]


@add_slots
@dataclass(frozen=True)
class ImportPartial:
    names: Sequence[ImportAlias]


@add_slots
@dataclass(frozen=True)
class ImportRelativePartial:
    relative: Sequence[Dot]
    module: Optional[Union[Attribute, Name]]


@add_slots
@dataclass(frozen=True)
class FormattedStringConversionPartial:
    value: str
    whitespace_before: WhitespaceState


@add_slots
@dataclass(frozen=True)
class FormattedStringFormatSpecPartial:
    values: Sequence[BaseFormattedStringContent]
    whitespace_before: WhitespaceState


@add_slots
@dataclass(frozen=True)
class ExceptClausePartial:
    leading_lines: Sequence[EmptyLine]
    whitespace_after_except: SimpleWhitespace
    type: Optional[BaseExpression] = None
    name: Optional[AsName] = None



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
from types import MappingProxyType
from typing import (
    TYPE_CHECKING,
    Callable,
    Generic,
    List,
    Mapping,
    MutableMapping,
    Optional,
    Type,
    TypeVar,
    cast,
)

from libcst._batched_visitor import BatchableCSTVisitor
from libcst._metadata_dependent import (
    _T as _MetadataT,
    _UNDEFINED_DEFAULT,
    MetadataDependent,
)
from libcst._visitors import CSTVisitor

if TYPE_CHECKING:
    from libcst._nodes.base import CSTNode
    from libcst._nodes.module import Module, _ModuleSelfT as _ModuleT
    from libcst.metadata.wrapper import MetadataWrapper


ProviderT = Type["BaseMetadataProvider[object]"]
# BaseMetadataProvider[int] would be a subtype of BaseMetadataProvider[object], so the
# typevar is covariant.
_ProvidedMetadataT = TypeVar("_ProvidedMetadataT", covariant=True)


# We can't use an ABCMeta here, because of metaclass conflicts
class BaseMetadataProvider(MetadataDependent, Generic[_ProvidedMetadataT]):
    """
    The low-level base class for all metadata providers. This class should be
    extended for metadata providers that are not visitor-based.

    This class is generic. A subclass of ``BaseMetadataProvider[T]`` will
    provider metadata of type ``T``.
    """

    #: Cache of metadata computed by this provider
    #
    # N.B. This has some typing variance problems. See `set_metadata` for an
    # explanation.
    _computed: MutableMapping["CSTNode", _ProvidedMetadataT]

    #: Implement gen_cache to indicate the matadata provider depends on cache from external
    #: system. This function will be called by :class:`~libcst.metadata.FullRepoManager`
    #: to compute required cache object per file path.
    gen_cache: Optional[Callable[[Path, List[str], int], Mapping[str, object]]] = None

    def __init__(self, cache: object = None) -> None:
        super().__init__()
        self._computed = {}
        if self.gen_cache and cache is None:
            # The metadata provider implementation is responsible to store and use cache.
            raise Exception(
                f"Cache is required for initializing {self.__class__.__name__}."
            )
        self.cache = cache

    def _gen(
        self, wrapper: "MetadataWrapper"
    ) -> Mapping["CSTNode", _ProvidedMetadataT]:
        """
        Resolves and returns metadata mapping for the module in ``wrapper``.

        This method is used by the metadata resolver and should not be called
        directly.
        """

        self._computed = {}
        # Resolve metadata dependencies for this provider
        with self.resolve(wrapper):
            self._gen_impl(wrapper.module)

        # Copy into a mapping proxy to ensure immutability
        return MappingProxyType(dict(self._computed))

    def _gen_impl(self, module: "Module") -> None:
        """
        Override this method with a metadata computation implementation.
        """
        ...

    # pyre-ignore[46]: The covariant `value` isn't type-safe because we write it to
    # pyre: `self._computed`, however we assume that only one subclass in the MRO chain
    # pyre: will ever call `set_metadata`, so it's okay for our purposes. There's no
    # pyre: sane way to redesign this API so that it doesn't have this problem.
    def set_metadata(self, node: "CSTNode", value: _ProvidedMetadataT) -> None:
        """
        Record a metadata value ``value`` for ``node``.
        """
        self._computed[node] = value

    def get_metadata(
        self,
        key: Type["BaseMetadataProvider[_MetadataT]"],
        node: "CSTNode",
        default: _MetadataT = _UNDEFINED_DEFAULT,
    ) -> _MetadataT:
        """
        The same method as :func:`~libcst.MetadataDependent.get_metadata` except
        metadata is accessed from ``self._computed`` in addition to ``self.metadata``.
        See :func:`~libcst.MetadataDependent.get_metadata`.
        """
        if key is type(self):
            if default is not _UNDEFINED_DEFAULT:
                return cast(_MetadataT, self._computed.get(node, default))
            else:
                return cast(_MetadataT, self._computed[node])

        return super().get_metadata(key, node, default)


class VisitorMetadataProvider(CSTVisitor, BaseMetadataProvider[_ProvidedMetadataT]):
    """
    The low-level base class for all non-batchable visitor-based metadata
    providers. Inherits from :class:`~libcst.CSTVisitor`.

    This class is generic. A subclass of ``VisitorMetadataProvider[T]`` will
    provider metadata of type ``T``.
    """

    def _gen_impl(self, module: "_ModuleT") -> None:
        module.visit(self)


class BatchableMetadataProvider(
    BatchableCSTVisitor, BaseMetadataProvider[_ProvidedMetadataT]
):
    """
    The low-level base class for all batchable visitor-based metadata providers.
    Batchable providers should be preferred when possible as they are more
    efficient to run compared to non-batchable visitor-based providers.
    Inherits from :class:`~libcst.BatchableCSTVisitor`.

    This class is generic. A subclass of ``BatchableMetadataProvider[T]`` will
    provider metadata of type ``T``.
    """

    def _gen_impl(self, module: "Module") -> None:
        """
        Batchables providers are resolved through _gen_batchable] so no
        implementation should be provided in _gen_impl.
        """
        pass



import pandas as pd
import io

csv = '''
breed,group,price,longevity,size
Labrador Retriever,sporting,810.0,12.04,medium
German Shepherd,herding,820.0,9.73,large
Beagle,hound,288.0,12.3,small
Golden Retriever,sporting,958.0,12.04,medium
Yorkshire Terrier,toy,1057.0,12.6,small
Bulldog,non-sporting,2680.0,6.29,medium
Boxer,working,700.0,8.81,medium
Poodle,non-sporting,900.0,11.95,medium
Dachshund,hound,423.0,12.63,small
Rottweiler,working,1118.0,9.11,large
Shih Tzu,toy,583.0,13.2,small
Miniature Schnauzer,terrier,715.0,11.81,small
Doberman Pinscher,working,790.0,10.33,large
Chihuahua,toy,588.0,16.5,small
German Shorthaired Pointer,sporting,545.0,11.46,large
Siberian Husky,working,650.0,12.58,medium
Pomeranian,toy,670.0,9.67,small
French Bulldog,non-sporting,1900.0,9.0,medium
Great Dane,working,1040.0,6.96,large
Shetland Sheepdog,herding,465.0,12.53,small
Cavalier King Charles Spaniel,toy,1017.0,11.29,small
Boston Terrier,non-sporting,690.0,10.92,medium
Maltese,toy,650.0,12.25,small
Australian Shepherd,herding,565.0,12.28,medium
Pembroke Welsh Corgi,herding,587.0,12.25,small
Pug,toy,469.0,11.0,medium
Cocker Spaniel,sporting,465.0,12.5,small
Mastiff,working,900.0,6.5,large
English Springer Spaniel,sporting,615.0,12.54,medium
Brittany,sporting,618.0,12.92,medium
Bernese Mountain Dog,working,1320.0,7.56,large
West Highland White Terrier,terrier,538.0,12.8,small
Papillon,toy,740.0,13.0,small
Bichon Frise,non-sporting,693.0,12.21,small
Bullmastiff,working,980.0,7.57,large
Basset Hound,hound,490.0,11.43,small
Newfoundland,working,1178.0,9.32,large
Rhodesian Ridgeback,hound,995.0,9.1,large
Border Collie,herding,623.0,12.52,medium
Chesapeake Bay Retriever,sporting,522.0,9.48,large
'''

dogs = pd.read_csv(io.StringIO(csv))

(dogs[dogs['size'] == 'small']
 .sort_values('group')
 .groupby('group').median()
)


# adapted from
# https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.iloc.html
# example where a boolean slice happens on the *columns*, not the rows

import pandas as pd

mydict = [{'a': 1, 'b': 2, 'c': 3, 'd': 4},
          {'a': 100, 'b': 200, 'c': 300, 'd': 400},
          {'a': 1000, 'b': 2000, 'c': 3000, 'd': 4000 }]

df = pd.DataFrame(mydict, index=['one', 'two', 'three'])

df.loc[:, df.loc['one'] % 2 == 0]


from pathlib import Path
import unittest
import json
from dataclasses import asdict

from pandas_tutor.parse import parse

test_cases = Path(__file__).parent / 'parse_golden'


class TestParse(unittest.TestCase):
    maxDiff = None
    pass


def make_test_case(test_name):
    in_file = test_cases / f'{test_name}.py'
    golden_file = test_cases / f'{test_name}.py.golden'

    def test(self: TestParse):
        self.assertTrue(in_file.exists())
        self.assertTrue(golden_file.exists())
        code = in_file.read_text()
        res = asdict(parse(code))
        golden_res = json.loads(golden_file.read_text())
        self.assertEqual(res, golden_res)

    return test


# make all test cases dynamically!
for test_name in test_cases.iterdir():
    if test_name.suffix == '.py' and test_name.name != '__init__.py':
        test_case = make_test_case(test_name.stem)
        setattr(TestParse, f'test_{test_name.stem}', test_case)
